# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from dataclasses import FrozenInstanceError

import pytest
import scipp as sc

import tof

Hz = sc.Unit('Hz')
deg = sc.Unit('deg')
meter = sc.Unit('m')
ms = sc.Unit('ms')
topen = 10.0 * ms
tclose = 20.0 * ms


@pytest.fixture
def chopper(make_chopper):
    return make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper',
    )


@pytest.fixture
def detector():
    return tof.Detector(distance=20 * meter, name='detector')


@pytest.fixture
def source(chopper, make_source):
    return make_source(
        arrival_times=sc.concat(
            [0.9 * topen, 0.5 * (topen + tclose), 1.1 * tclose], dim='event'
        ),
        distance=chopper.distance,
    )


@pytest.fixture
def multi_pulse_source(chopper, make_source):
    return make_source(
        arrival_times=sc.concat(
            [0.9 * topen, 0.5 * (topen + tclose), 1.1 * tclose], dim='event'
        ),
        distance=chopper.distance,
        pulses=3,
        frequency=14.0 * Hz,
    )


@pytest.fixture
def model(chopper, detector, source):
    return tof.Model(source=source, choppers=[chopper], detectors=[detector])


def make_ess_model(pulses=1, frequency=None):
    source = tof.Source(facility='ess', neutrons=100_000, pulses=pulses)
    detector = tof.Detector(distance=30.0 * meter, name='detector')
    if frequency is None:
        frequency = 14.0 * Hz
    chopper = tof.Chopper(
        frequency=frequency,
        open=sc.array(
            dims=['cutout'],
            values=[30.0, 50.0],
            unit='deg',
        ),
        close=sc.array(
            dims=['cutout'],
            values=[40.0, 80.0],
            unit='deg',
        ),
        phase=0.0 * deg,
        distance=8 * meter,
        name='chopper',
    )
    return tof.Model(source=source, choppers=[chopper], detectors=[detector])


def test_source_results_are_read_only(source, model):
    res = model.run()

    # Check that basic properties are accessible
    assert res.source.neutrons == 3
    assert sc.identical(
        res.source.data['pulse', 0].coords['wavelength'],
        source.data['pulse', 0].coords['wavelength'],
    )
    # Check that values cannot be overwritten
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        res.source.neutrons = 5
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        res.source.data = 'corrupted'


def test_chopper_results_are_read_only(chopper, model):
    res = model.run()
    ch = res.choppers['chopper']

    # Check that basic properties are accessible
    assert ch.toa.data.sum().value == 1
    assert ch.toa.data.masks['blocked_by_me'].sum().value == 2
    assert sc.identical(ch.distance, chopper.distance)
    assert sc.identical(ch.phase, chopper.phase)
    # Check that values cannot be overwritten
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        ch.phase = 63.0 * deg
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        ch.frequency = 21.0 * Hz
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        ch.toa = [1, 2, 3, 4, 5]
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        ch.toa.data = [1, 2, 3, 4, 5]
    # Check that choppers cannot be added or removed
    with pytest.raises(TypeError, match='object does not support item assignment'):
        res.choppers['chopper2'] = chopper
    with pytest.raises(TypeError, match='object does not support item deletion'):
        del res.choppers['chopper']


def test_detector_results_are_read_only(detector, model):
    res = model.run()
    det = res.detectors['detector']

    # Check that basic properties are accessible
    assert det.toa.data.sum().value == 1
    assert sc.identical(det.distance, detector.distance)
    # Check that values cannot be overwritten
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        det.distance = 55.0 * meter
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        det.wavelength = sc.arange('wavelength', 20.0, unit='angstrom')
    with pytest.raises(FrozenInstanceError, match='cannot assign to field'):
        det.wavelength.data = sc.arange('wavelength', 20.0, unit='angstrom')
    # Check that detectors cannot be added or removed
    with pytest.raises(TypeError, match='object does not support item assignment'):
        res.detectors['detector2'] = detector
    with pytest.raises(TypeError, match='object does not support item deletion'):
        del res.detectors['detector']


def test_component_results_data_access(chopper, detector, multi_pulse_source):
    model = tof.Model(
        source=multi_pulse_source, choppers=[chopper], detectors=[detector]
    )
    res = model.run()
    ch = res.choppers['chopper']
    det = res.detectors['detector']

    for field in ('toa', 'wavelength', 'birth_time', 'speed'):
        assert field in getattr(ch, field).data.coords
        assert field in getattr(det, field).data.coords
        assert 'blocked_by_me' in getattr(ch, field).data.masks
        assert 'blocked_by_others' in getattr(ch, field).data.masks
        assert 'blocked_by_others' in getattr(det, field).data.masks
        assert getattr(ch, field).data.sizes['pulse'] == 3
        assert getattr(det, field).data.sizes['pulse'] == 3


def test_component_data_slicing(chopper, detector, multi_pulse_source):
    model = tof.Model(
        source=multi_pulse_source, choppers=[chopper], detectors=[detector]
    )
    res = model.run()
    ch = res.choppers['chopper']

    toas = ch.toa['pulse', 0].data
    assert 'pulse' not in toas.dims
    assert 'event' in toas.dims


def test_component_results_data_slice_range(chopper, detector, multi_pulse_source):
    model = tof.Model(
        source=multi_pulse_source, choppers=[chopper], detectors=[detector]
    )
    res = model.run()
    ch = res.choppers['chopper']

    toas = ch.toa['pulse', 0:2].data
    assert toas.sizes['pulse'] == 2
    assert 'event' in toas.dims


def test_component_results_data_slice_negative_index(
    chopper, detector, multi_pulse_source
):
    model = tof.Model(
        source=multi_pulse_source, choppers=[chopper], detectors=[detector]
    )
    res = model.run()
    ch = res.choppers['chopper']

    toas = ch.toa['pulse', 0:-1].data
    assert toas.sizes['pulse'] == 2
    assert 'event' in toas.dims


def test_component_results_data_slice_step(chopper, detector, multi_pulse_source):
    model = tof.Model(
        source=multi_pulse_source, choppers=[chopper], detectors=[detector]
    )
    res = model.run()
    ch = res.choppers['chopper']

    toas = ch.toa['pulse', ::2].data
    assert toas.sizes['pulse'] == 2
    assert 'event' in toas.dims


def test_component_results_slice_all_results(chopper, detector, multi_pulse_source):
    model = tof.Model(
        source=multi_pulse_source, choppers=[chopper], detectors=[detector]
    )
    res = model.run()
    ch = res.choppers['chopper']

    reading = ch['pulse', 0]
    toas = reading.toa.data
    assert 'pulse' not in toas.dims
    assert 'event' in toas.dims
    wavelengths = reading.wavelength.data
    assert 'pulse' not in wavelengths.dims
    assert 'event' in wavelengths.dims


def test_component_results_slice_no_pulse_dim(chopper, detector, multi_pulse_source):
    model = tof.Model(
        source=multi_pulse_source, choppers=[chopper], detectors=[detector]
    )
    res = model.run()
    ch = res.choppers['chopper']

    reading = ch[0]  # No pulse dim
    data = reading.data
    assert 'pulse' not in data.dims
    assert 'event' in data.dims

    toas = ch.toa[1].data
    assert 'pulse' not in toas.dims
    assert 'event' in toas.dims


def test_component_results_eto(chopper, detector, multi_pulse_source):
    model = tof.Model(
        source=multi_pulse_source, choppers=[chopper], detectors=[detector]
    )
    res = model.run()
    data = res['detector'].data
    assert 'eto' in data.coords
    assert data.coords['eto'].max() < (1 / model.source.frequency).to(
        unit=data.coords['eto'].unit
    )


def test_result_plot_does_not_raise():
    model = make_ess_model()
    res = model.run()
    res.plot()
    res.plot(visible_rays=5000)
    res.plot(visible_rays=50, blocked_rays=3000)


def test_result_multiple_pulses_plot_does_not_raise():
    model = make_ess_model(pulses=3)
    res = model.run()
    res.plot()
    res.plot(visible_rays=2000)
    res.plot(visible_rays=50, blocked_rays=3000)


def test_result_multiple_pulses_plot_with_no_events_in_last_frame_does_not_raise():
    model = make_ess_model(pulses=3, frequency=10.0 * Hz)
    res = model.run()
    res.plot()
    res.plot(visible_rays=2000)
    res.plot(visible_rays=50, blocked_rays=3000)


def test_result_all_neutrons_blocked_does_not_raise():
    model = make_ess_model(frequency=1.0 * Hz)
    res = model.run()
    res.plot()
    res.plot(blocked_rays=500)


def test_result_plot_cmap_does_not_raise():
    model = make_ess_model()
    res = model.run()
    res.plot(cmap='viridis')


def test_result_plot_no_detectors_does_not_raise():
    model = make_ess_model()
    model._detectors = {}
    res = model.run()
    res.plot()
    res.plot(visible_rays=5000)
    res.plot(visible_rays=50, blocked_rays=3000)


def test_result_repr_does_not_raise():
    model = make_ess_model()
    res = model.run()
    assert repr(res) is not None


def test_component_repr_does_not_raise():
    model = make_ess_model()
    res = model.run()
    assert repr(res.choppers['chopper']) is not None
    assert repr(res.detectors['detector']) is not None


def test_componentdata_repr_does_not_raise():
    model = make_ess_model()
    res = model.run()
    assert repr(res.choppers['chopper'].toa) is not None
    assert repr(res.detectors['detector'].wavelength) is not None


def test_plot_reading_does_not_raise(model):
    res = model.run()
    res.choppers['chopper'].toa.plot()
    res.choppers['chopper'].wavelength.plot()
    res.detectors['detector'].toa.plot()
    res.detectors['detector'].wavelength.plot()


def test_plot_reading_pulse_skipping_does_not_raise():
    model = make_ess_model(pulses=3)
    skip = tof.Chopper(
        frequency=7 * Hz,
        open=sc.array(dims=['cutout'], values=[0.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[180.0], unit='deg'),
        phase=0.0 * deg,
        distance=10 * meter,
        name='skip',
    )
    model.choppers['skip'] = skip
    res = model.run()
    res.choppers['chopper'].toa.plot()
    res.choppers['chopper'].wavelength.plot()
    res.detectors['detector'].toa.plot()
    res.detectors['detector'].wavelength.plot()


def test_plot_reading_nothing_to_plot_raises():
    model = make_ess_model(pulses=1)
    skip = tof.Chopper(
        frequency=1 * Hz,
        open=sc.array(dims=['cutout'], values=[0.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[1.0], unit='deg'),
        phase=0.0 * deg,
        distance=10 * meter,
        name='skip',
    )
    model.choppers['skip'] = skip
    res = model.run()
    with pytest.raises(RuntimeError, match="Nothing to plot."):
        res.detectors['detector'].toa.plot()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc
from scippneutron.chopper import DiskChopper, DiskChopperType

import tof

Hz = sc.Unit('Hz')
deg = sc.Unit('deg')
rad = sc.Unit('rad')
meter = sc.Unit('m')
sec = sc.Unit('s')
two_pi = 2.0 * np.pi * rad


def test_angular_speed():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=0.0 * deg,
        close=10.0 * deg,
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    assert chopper.omega == two_pi * f


def test_open_close_times_one_rotation():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0], unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )

    topen, tclose = chopper.open_close_times(0 * sec)
    # Note that choppers perform one rotation before t=0 to make sure
    # no openings are missed at early times.
    assert sc.identical(topen[0], ((10.0 * deg).to(unit='rad') - two_pi) / (two_pi * f))
    assert sc.identical(
        tclose[0], ((20.0 * deg).to(unit='rad') - two_pi) / (two_pi * f)
    )
    assert sc.identical(topen[1], (10.0 * deg).to(unit='rad') / (two_pi * f))
    assert sc.identical(tclose[1], (20.0 * deg).to(unit='rad') / (two_pi * f))


def test_open_close_times_three_rotations():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0], unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )

    topen, tclose = chopper.open_close_times(0.21 * sec)
    open0 = (10.0 * deg).to(unit='rad')
    close0 = (20.0 * deg).to(unit='rad')
    assert sc.identical(
        topen,
        sc.concat(
            [open0 - two_pi, open0, open0 + two_pi, open0 + 2 * two_pi], dim='cutout'
        )
        / (two_pi * f),
    )
    assert sc.identical(
        tclose,
        sc.concat(
            [close0 - two_pi, close0, close0 + two_pi, close0 + 2 * two_pi],
            dim='cutout',
        )
        / (two_pi * f),
    )


def test_open_close_angles_scalars_converted_to_arrays():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=10.0 * deg,
        close=20.0 * deg,
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    topen, tclose = chopper.open_close_times(0.0 * sec)
    assert sc.identical(topen[1], (10.0 * deg).to(unit='rad') / (two_pi * f))
    assert sc.identical(tclose[1], (20.0 * deg).to(unit='rad') / (two_pi * f))


def test_phase():
    f = 10.0 * Hz
    chopper1 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0], unit='deg'),
        phase=0.0 * deg,
        distance=10.0 * meter,
        name='chopper1',
    )
    topen1, tclose1 = chopper1.open_close_times(0.0 * sec)
    chopper2 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0], unit='deg'),
        phase=30.0 * deg,
        distance=10.0 * meter,
        name='chopper2',
    )
    topen2, tclose2 = chopper2.open_close_times(0.0 * sec)
    assert sc.allclose(topen2, topen1 + (30.0 * deg).to(unit='rad') / chopper2.omega)
    assert sc.allclose(tclose2, tclose1 + (30.0 * deg).to(unit='rad') / chopper2.omega)


def test_phase_int():
    f = 10.0 * Hz
    op = sc.array(dims=['cutout'], values=[10.0], unit='deg')
    cl = sc.array(dims=['cutout'], values=[20.0], unit='deg')
    d = 10.0 * meter
    chopper1 = tof.Chopper(
        frequency=f,
        open=op,
        close=cl,
        phase=30.0 * deg,
        distance=d,
        name='chopper1',
    )
    chopper2 = tof.Chopper(
        frequency=f,
        open=op,
        close=cl,
        phase=30 * deg,
        distance=d,
        name='chopper2',
    )
    topen1, tclose1 = chopper1.open_close_times(0.0 * sec)
    topen2, tclose2 = chopper2.open_close_times(0.0 * sec)
    assert sc.allclose(topen1, topen2)
    assert sc.allclose(tclose1, tclose2)


def test_frequency_must_be_positive():
    with pytest.raises(ValueError, match="Chopper frequency must be positive"):
        tof.Chopper(
            frequency=-1.0 * Hz,
            open=0.0 * deg,
            close=10.0 * deg,
            phase=0.0 * deg,
            distance=5.0 * meter,
            name='chopper',
        )


def test_open_close_times_counter_rotation():
    f = 10.0 * Hz
    d = 10.0 * meter
    ph = 0.0 * deg
    chopper1 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0, 90.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0, 130.0], unit='deg'),
        phase=ph,
        distance=d,
        name='chopper1',
    )
    chopper2 = tof.Chopper(
        frequency=f,
        open=sc.array(
            dims=['cutout'], values=[360.0 - 130.0, 360.0 - 20.0], unit='deg'
        ),
        close=sc.array(
            dims=['cutout'], values=[360.0 - 90.0, 360.0 - 10.0], unit='deg'
        ),
        phase=ph,
        distance=d,
        direction=tof.AntiClockwise,
        name='chopper2',
    )

    topen1, tclose1 = chopper1.open_close_times(0.0 * sec)
    topen2, tclose2 = chopper2.open_close_times(0.0 * sec)
    assert sc.allclose(topen1, topen2)
    assert sc.allclose(tclose1, tclose2)


def test_open_close_times_counter_rotation_with_phase():
    f = 10.0 * Hz
    chopper1 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[80.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[90.0], unit='deg'),
        phase=0.0 * deg,
        distance=10.0 * meter,
        direction=tof.AntiClockwise,
        name='chopper1',
    )
    topen1, tclose1 = chopper1.open_close_times(0.0 * sec)
    chopper2 = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[80.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[90.0], unit='deg'),
        phase=30.0 * deg,
        distance=10.0 * meter,
        direction=tof.AntiClockwise,
        name='chopper2',
    )
    topen2, tclose2 = chopper2.open_close_times(0.0 * sec)
    assert sc.allclose(
        topen2, topen1 + (30.0 * deg).to(unit='rad') / abs(chopper2.omega)
    )
    assert sc.allclose(
        tclose2, tclose1 + (30.0 * deg).to(unit='rad') / abs(chopper2.omega)
    )


def test_open_close_anticlockwise_multiple_rotations():
    chopper = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.array(dims=["cutout"], values=[10.0, 90.0], unit="deg"),
        close=sc.array(dims=["cutout"], values=[20.0, 130.0], unit="deg"),
        distance=10.0 * meter,
        phase=0 * deg,
        direction=tof.AntiClockwise,
        name='chopper',
    )

    two_rotations_open, two_rotations_close = chopper.open_close_times(0.0 * sec)
    three_rotations_open, three_rotations_close = chopper.open_close_times(0.2 * sec)
    four_rotations_open, four_rotations_close = chopper.open_close_times(0.3 * sec)

    assert len(two_rotations_open) == 4
    assert len(three_rotations_open) == 6
    assert len(four_rotations_open) == 8
    assert len(two_rotations_close) == 4
    assert len(three_rotations_close) == 6
    assert len(four_rotations_close) == 8

    assert sc.allclose(two_rotations_open, three_rotations_open[:-2])
    assert sc.allclose(two_rotations_close, three_rotations_close[:-2])
    assert sc.allclose(three_rotations_open, four_rotations_open[:-2])
    assert sc.allclose(three_rotations_close, four_rotations_close[:-2])
    assert sc.allclose(two_rotations_open, four_rotations_open[:-4])
    assert sc.allclose(two_rotations_close, four_rotations_close[:-4])


def test_bad_direction_raises():
    f = 10.0 * Hz
    op = sc.array(dims=['cutout'], values=[10.0], unit='deg')
    cl = sc.array(dims=['cutout'], values=[20.0], unit='deg')
    d = 10.0 * meter
    ph = 0.0 * deg
    tof.Chopper(
        frequency=f,
        open=op,
        close=cl,
        phase=ph,
        distance=d,
        direction=tof.Clockwise,
        name='chopper',
    )
    tof.Chopper(
        frequency=f,
        open=op,
        close=cl,
        phase=ph,
        distance=d,
        direction=tof.AntiClockwise,
        name='chopper',
    )
    with pytest.raises(
        ValueError, match="Chopper direction must be Clockwise or AntiClockwise"
    ):
        tof.Chopper(
            frequency=f,
            open=op,
            close=cl,
            phase=ph,
            distance=d,
            direction='clockwise',
            name='chopper',
        )
    with pytest.raises(
        ValueError, match="Chopper direction must be Clockwise or AntiClockwise"
    ):
        tof.Chopper(
            frequency=f,
            open=op,
            close=cl,
            phase=ph,
            distance=d,
            direction='anti-clockwise',
            name='chopper',
        )
    with pytest.raises(
        ValueError, match="Chopper direction must be Clockwise or AntiClockwise"
    ):
        tof.Chopper(
            frequency=f,
            open=op,
            close=cl,
            phase=ph,
            distance=d,
            direction=1,
            name='chopper',
        )


def test_chopper_create_from_centers_and_widths():
    f = 10.0 * Hz
    centers = sc.array(dims=['cutout'], values=[15.0, 46.0], unit='deg')
    widths = sc.array(dims=['cutout'], values=[10.0, 16.0], unit='deg')
    chopper = tof.Chopper(
        frequency=f,
        centers=centers,
        widths=widths,
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    assert sc.allclose(chopper.open, centers - widths / 2)
    assert sc.allclose(chopper.close, centers + widths / 2)

    expected = tof.Chopper(
        frequency=f,
        open=sc.array(dims=['cutout'], values=[10.0, 38.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[20.0, 54.0], unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    assert sc.allclose(chopper.open, expected.open)
    assert sc.allclose(chopper.close, expected.close)


def test_chopper_create_raises_when_both_edges_and_centers_are_supplied():
    with pytest.raises(
        ValueError, match="Either open/close or centers/widths must be provided"
    ):
        tof.Chopper(
            frequency=10.0 * Hz,
            open=10.0 * deg,
            close=20.0 * deg,
            centers=15.0 * deg,
            widths=10.0 * deg,
            phase=0.0 * deg,
            distance=5.0 * meter,
            name='chopper',
        )


def test_as_json():
    f = 10.0 * Hz
    chopper = tof.Chopper(
        frequency=f,
        open=7.7 * deg,
        close=31.0 * deg,
        phase=0.5 * deg,
        distance=5.0 * meter,
        name='Achopper',
    )
    json_str = chopper.as_json()
    assert json_str['type'] == 'chopper'
    assert json_str['frequency']['value'] == chopper.frequency.value
    assert json_str['frequency']['unit'] == chopper.frequency.unit
    assert np.allclose(json_str['open']['value'], chopper.open.values)
    assert json_str['open']['unit'] == chopper.open.unit
    assert np.allclose(json_str['close']['value'], chopper.close.values)
    assert json_str['close']['unit'] == chopper.close.unit
    assert json_str['phase']['value'] == chopper.phase.value
    assert json_str['phase']['unit'] == chopper.phase.unit
    assert json_str['distance']['value'] == chopper.distance.value
    assert json_str['distance']['unit'] == chopper.distance.unit
    assert json_str['name'] == chopper.name
    assert json_str['direction'] == chopper.direction.name.lower()


def test_equal():
    chop = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.arange('cutout', 0, 180, 20.0, unit='deg'),
        close=sc.arange('cutout', 10, 190, 20.0, unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper',
    )
    assert chop == chop


@pytest.mark.parametrize("attr", ["frequency", "open", "close", "phase", "distance"])
def test_not_equal(attr):
    chop1 = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.arange('cutout', 0, 180, 20.0, unit='deg'),
        close=sc.arange('cutout', 10, 190, 20.0, unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper1',
    )
    chop2 = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.arange('cutout', 0, 180, 20.0, unit='deg'),
        close=sc.arange('cutout', 10, 190, 20.0, unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper1',
    )
    setattr(
        chop2,
        attr,
        getattr(chop1, attr) + sc.scalar(1.0, unit=getattr(chop1, attr).unit),
    )
    assert chop1 != chop2


def test_not_equal_different_name():
    chop1 = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.arange('cutout', 0, 180, 20.0, unit='deg'),
        close=sc.arange('cutout', 10, 190, 20.0, unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper1',
    )
    chop2 = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.arange('cutout', 0, 180, 20.0, unit='deg'),
        close=sc.arange('cutout', 10, 190, 20.0, unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper2',
    )
    assert chop1 != chop2


@pytest.mark.parametrize("attr", ["frequency", "open", "close", "phase", "distance"])
def test_not_equal_different_unit(attr):
    chop1 = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.arange('cutout', 0, 180, 20.0, unit='deg'),
        close=sc.arange('cutout', 10, 190, 20.0, unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper1',
    )
    chop2 = tof.Chopper(
        frequency=10.0 * Hz,
        open=sc.arange('cutout', 0, 180, 20.0, unit='deg'),
        close=sc.arange('cutout', 10, 190, 20.0, unit='deg'),
        phase=0.0 * deg,
        distance=5.0 * meter,
        name='chopper1',
    )
    if attr == "frequency":
        setattr(
            chop2,
            attr,
            getattr(chop1, attr).to(unit='kHz'),
        )
    elif attr == "distance":
        setattr(
            chop2,
            attr,
            getattr(chop1, attr).to(unit='cm'),
        )
    elif attr in ["open", "close", "phase"]:
        setattr(
            chop2,
            attr,
            getattr(chop1, attr).to(unit='rad'),
        )
    assert chop1 != chop2


def test_from_diskchopper():
    disk_chopper = DiskChopper(
        axle_position=sc.vector([0.0, 0.0, 2.0], unit='m'),
        frequency=sc.scalar(12.0, unit='Hz'),
        beam_position=sc.scalar(45.0, unit='deg'),
        phase=sc.scalar(20.0, unit='deg'),
        slit_begin=sc.array(dims=['slit'], values=[0.0, 124.0], unit='deg'),
        slit_end=sc.array(dims=['slit'], values=[60.0, 126.0], unit='deg'),
    )

    chopper = tof.Chopper.from_diskchopper(disk_chopper)

    assert sc.identical(chopper.frequency, sc.abs(disk_chopper.frequency))
    assert sc.identical(chopper.distance, disk_chopper.axle_position.fields.z)
    assert sc.identical(chopper.phase, disk_chopper.phase)
    assert sc.allclose(
        chopper.open, disk_chopper.slit_begin - disk_chopper.beam_position
    )
    assert sc.allclose(
        chopper.close, disk_chopper.slit_end - disk_chopper.beam_position
    )
    assert chopper.direction == tof.AntiClockwise


def test_from_diskchopper_clockwise():
    disk_chopper = DiskChopper(
        axle_position=sc.vector([0.0, 0.0, 2.0], unit='m'),
        frequency=sc.scalar(-12.0, unit='Hz'),
        beam_position=sc.scalar(45.0, unit='deg'),
        phase=sc.scalar(-20.0, unit='deg'),
        slit_begin=sc.array(dims=['slit'], values=[0.0, 124.0], unit='deg'),
        slit_end=sc.array(dims=['slit'], values=[60.0, 126.0], unit='deg'),
    )

    chopper = tof.Chopper.from_diskchopper(disk_chopper)

    assert sc.identical(chopper.frequency, sc.abs(disk_chopper.frequency))
    assert sc.identical(chopper.distance, disk_chopper.axle_position.fields.z)
    assert sc.identical(chopper.phase, -disk_chopper.phase)
    assert sc.allclose(
        chopper.open, disk_chopper.slit_begin - disk_chopper.beam_position
    )
    assert sc.allclose(
        chopper.close, disk_chopper.slit_end - disk_chopper.beam_position
    )
    assert chopper.direction == tof.Clockwise


def test_from_diskchopper_with_name():
    disk_chopper = DiskChopper(
        axle_position=sc.vector([0.0, 0.0, 2.0], unit='m'),
        frequency=sc.scalar(12.0, unit='Hz'),
        beam_position=sc.scalar(45.0, unit='deg'),
        phase=sc.scalar(20.0, unit='deg'),
        slit_begin=sc.array(dims=['slit'], values=[0.0, 124.0], unit='deg'),
        slit_end=sc.array(dims=['slit'], values=[60.0, 126.0], unit='deg'),
    )

    chopper = tof.Chopper.from_diskchopper(disk_chopper, name="TestChopper")
    assert chopper.name == "TestChopper"


def test_to_diskchopper():
    chopper = tof.Chopper(
        frequency=12.0 * Hz,
        open=sc.array(dims=['cutout'], values=[-15.0, 45.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[15.0, 75.0], unit='deg'),
        phase=20.0 * deg,
        distance=2.0 * meter,
        name='chopper',
        direction=tof.AntiClockwise,
    )

    disk_chopper = chopper.to_diskchopper()

    assert sc.identical(disk_chopper.frequency, chopper.frequency)
    assert sc.identical(
        disk_chopper.axle_position,
        sc.vector(value=[0.0, 0.0, chopper.distance.value], unit=chopper.distance.unit),
    )
    assert sc.identical(disk_chopper.phase, chopper.phase)
    assert sc.allclose(disk_chopper.slit_begin, chopper.open)
    assert sc.allclose(disk_chopper.slit_end, chopper.close)
    assert sc.identical(disk_chopper.beam_position, sc.scalar(0.0, unit='deg'))


def test_diskchopper_roundtrip():
    original_disk_chopper = DiskChopper(
        axle_position=sc.vector([0.0, 0.0, 2.0], unit='m'),
        frequency=sc.scalar(12.0, unit='Hz'),
        beam_position=sc.scalar(0.0, unit='deg'),
        phase=sc.scalar(20.0, unit='deg'),
        slit_begin=sc.array(dims=['slit'], values=[0.0, 124.0], unit='deg'),
        slit_end=sc.array(dims=['slit'], values=[60.0, 126.0], unit='deg'),
    )

    chopper = tof.Chopper.from_diskchopper(original_disk_chopper)
    converted_disk_chopper = chopper.to_diskchopper()

    assert sc.identical(
        original_disk_chopper.frequency, converted_disk_chopper.frequency
    )
    assert sc.identical(
        original_disk_chopper.axle_position, converted_disk_chopper.axle_position
    )
    assert sc.identical(original_disk_chopper.phase, converted_disk_chopper.phase)
    assert sc.allclose(
        original_disk_chopper.slit_begin, converted_disk_chopper.slit_begin
    )
    assert sc.allclose(original_disk_chopper.slit_end, converted_disk_chopper.slit_end)
    assert sc.identical(
        original_disk_chopper.beam_position, converted_disk_chopper.beam_position
    )


def test_diskchopper_roundtrip_with_beamposition():
    original_disk_chopper = DiskChopper(
        axle_position=sc.vector([0.0, 0.0, 2.0], unit='m'),
        frequency=sc.scalar(12.0, unit='Hz'),
        beam_position=sc.scalar(45.0, unit='deg'),
        phase=sc.scalar(20.0, unit='deg'),
        slit_begin=sc.array(dims=['slit'], values=[0.0, 124.0], unit='deg'),
        slit_end=sc.array(dims=['slit'], values=[60.0, 126.0], unit='deg'),
    )

    chopper = tof.Chopper.from_diskchopper(original_disk_chopper)
    converted_disk_chopper = chopper.to_diskchopper()

    assert sc.identical(
        original_disk_chopper.frequency, converted_disk_chopper.frequency
    )
    assert sc.identical(
        original_disk_chopper.axle_position, converted_disk_chopper.axle_position
    )
    assert sc.identical(original_disk_chopper.phase, converted_disk_chopper.phase)
    assert sc.allclose(
        original_disk_chopper.slit_begin,
        converted_disk_chopper.slit_begin + original_disk_chopper.beam_position,
    )
    assert sc.allclose(
        original_disk_chopper.slit_end,
        converted_disk_chopper.slit_end + original_disk_chopper.beam_position,
    )
    assert sc.identical(
        converted_disk_chopper.beam_position, sc.scalar(0.0, unit='deg')
    )


def test_from_nexus():
    nexus_chopper = {
        'type': DiskChopperType.single,
        'position': sc.vector([0.0, 0.0, 2.0], unit='m'),
        'rotation_speed': sc.scalar(12.0, unit='Hz'),
        'beam_position': sc.scalar(45.0, unit='deg'),
        'phase': sc.scalar(20.0, unit='deg'),
        'slit_edges': sc.array(
            dims=['slit'],
            values=[0.0, 60.0, 124.0, 126.0],
            unit='deg',
        ),
        'slit_height': sc.scalar(0.4, unit='m'),
        'radius': sc.scalar(0.5, unit='m'),
    }

    chopper = tof.Chopper.from_nexus(nexus_chopper)

    assert sc.identical(chopper.frequency, sc.abs(nexus_chopper['rotation_speed']))
    assert sc.identical(chopper.distance, nexus_chopper['position'].fields.z)
    assert sc.identical(chopper.phase, nexus_chopper['phase'])
    assert sc.allclose(
        chopper.open, nexus_chopper['slit_edges'][::2] - nexus_chopper['beam_position']
    )
    assert sc.allclose(
        chopper.close,
        nexus_chopper['slit_edges'][1::2] - nexus_chopper['beam_position'],
    )
    assert chopper.direction == tof.AntiClockwise


def test_from_nexus_clockwise():
    nexus_chopper = {
        'type': DiskChopperType.single,
        'position': sc.vector([0.0, 0.0, 2.0], unit='m'),
        'rotation_speed': sc.scalar(-12.0, unit='Hz'),
        'beam_position': sc.scalar(45.0, unit='deg'),
        'phase': sc.scalar(-20.0, unit='deg'),
        'slit_edges': sc.array(
            dims=['slit'],
            values=[0.0, 60.0, 124.0, 126.0],
            unit='deg',
        ),
        'slit_height': sc.scalar(0.4, unit='m'),
        'radius': sc.scalar(0.5, unit='m'),
    }

    chopper = tof.Chopper.from_nexus(nexus_chopper)

    assert sc.identical(chopper.frequency, sc.abs(nexus_chopper['rotation_speed']))
    assert sc.identical(chopper.distance, nexus_chopper['position'].fields.z)
    assert sc.identical(chopper.phase, -nexus_chopper['phase'])
    assert sc.allclose(
        chopper.open, nexus_chopper['slit_edges'][::2] - nexus_chopper['beam_position']
    )
    assert sc.allclose(
        chopper.close,
        nexus_chopper['slit_edges'][1::2] - nexus_chopper['beam_position'],
    )
    assert chopper.direction == tof.Clockwise



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

from itertools import chain
from types import MappingProxyType

import matplotlib.pyplot as plt
import numpy as np
import scipp as sc
from matplotlib.collections import LineCollection

from .chopper import Chopper, ChopperReading
from .detector import Detector, DetectorReading
from .source import Source, SourceParameters
from .utils import Plot, one_mask


def _add_rays(
    ax: plt.Axes,
    x: np.ndarray,
    y: np.ndarray,
    color: np.ndarray | str,
    cbar: bool = True,
    cmap: str = "gist_rainbow_r",
    vmin: float | None = None,
    vmax: float | None = None,
    cax: plt.Axes | None = None,
    zorder: int = 1,
):
    coll = LineCollection(np.stack((x, y), axis=2), zorder=zorder)
    if isinstance(color, str):
        coll.set_color(color)
    else:
        coll.set_cmap(plt.colormaps[cmap])
        coll.set_array(color)
        coll.set_norm(plt.Normalize(vmin, vmax))
        if cbar:
            cb = plt.colorbar(coll, ax=ax, cax=cax)
            cb.ax.yaxis.set_label_coords(-0.9, 0.5)
            cb.set_label("Wavelength [Å]")
    ax.add_collection(coll)


class Result:
    """
    Result of a simulation.

    Parameters
    ----------
    source:
        The source of neutrons.
    choppers:
        The choppers in the model.
    detectors:
        The detectors in the model.
    """

    def __init__(
        self,
        source: Source,
        choppers: dict[str, Chopper],
        detectors: dict[str, Detector],
    ):
        self._source = source.as_readonly()
        self._choppers = {}
        for name, chopper in choppers.items():
            self._choppers[name] = ChopperReading(
                distance=chopper["distance"],
                name=chopper["name"],
                frequency=chopper["frequency"],
                open=chopper["open"],
                close=chopper["close"],
                phase=chopper["phase"],
                open_times=chopper["open_times"],
                close_times=chopper["close_times"],
                data=chopper["data"],
            )

        self._detectors = {}
        for name, det in detectors.items():
            self._detectors[name] = DetectorReading(
                distance=det["distance"], name=det["name"], data=det["data"]
            )

        self._choppers = MappingProxyType(self._choppers)
        self._detectors = MappingProxyType(self._detectors)

    @property
    def choppers(self) -> MappingProxyType[str, ChopperReading]:
        """The choppers in the model."""
        return self._choppers

    @property
    def detectors(self) -> MappingProxyType[str, DetectorReading]:
        """The detectors in the model."""
        return self._detectors

    @property
    def source(self) -> SourceParameters:
        """The source of neutrons."""
        return self._source

    def __iter__(self):
        return chain(self._choppers, self._detectors)

    def __getitem__(self, name: str) -> ChopperReading | DetectorReading:
        if name not in self:
            raise KeyError(f"No component with name {name} was found.")
        return self._choppers[name] if name in self._choppers else self._detectors[name]

    def plot(
        self,
        visible_rays: int = 1000,
        blocked_rays: int = 0,
        figsize: tuple[float, float] | None = None,
        ax: plt.Axes | None = None,
        cax: plt.Axes | None = None,
        cbar: bool = True,
        cmap: str = "gist_rainbow_r",
        seed: int | None = None,
        vmin: float | None = None,
        vmax: float | None = None,
    ) -> Plot:
        """
        Plot the time-distance diagram for the instrument, including the rays of
        neutrons that make it to the furthest detector.
        As plotting many lines can be slow, the number of rays to plot can be
        limited by setting ``visible_rays``.
        In addition, it is possible to also plot the rays that are blocked by
        choppers along the flight path by setting ``blocked_rays > 0``.

        Parameters
        ----------
        visible_rays:
            Maximum number of rays to plot.
        blocked_rays:
            Number of blocked rays to plot.
        figsize:
            Figure size.
        ax:
            Axes to plot on.
        cax:
            Axes to use for the colorbar.
        cbar:
            Show a colorbar for the wavelength if ``True``.
        cmap:
            Colormap to use for the wavelength colorbar.
        seed:
            Random seed for reproducibility.
        vmin:
            Minimum value for the colorbar.
        vmax:
            Maximum value for the colorbar.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = ax.get_figure()
        components = sorted(
            chain(self.choppers.values(), self.detectors.values()),
            key=lambda c: c.distance,
        )
        furthest_component = components[-1]
        source_dist = self.source.distance.value
        repeats = [1] + [2] * len(components)

        wavelengths = sc.DataArray(
            data=furthest_component.data.coords["wavelength"],
            masks=furthest_component.data.masks,
        )
        wmin, wmax = wavelengths.min(), wavelengths.max()

        rng = np.random.default_rng(seed)

        for i in range(self._source.data.sizes["pulse"]):
            source_data = self.source.data["pulse", i]
            component_data = furthest_component.data["pulse", i]
            ids = np.arange(self.source.neutrons)
            # Plot visible rays
            blocked = one_mask(component_data.masks).values
            nblocked = int(blocked.sum())
            if nblocked < self.source.neutrons:
                inds = rng.choice(
                    ids[~blocked],
                    size=min(self.source.neutrons - nblocked, visible_rays),
                    replace=False,
                )

                xstart = source_data.coords["birth_time"].values[inds]
                xend = component_data.coords["toa"].values[inds]
                ystart = np.full_like(xstart, source_dist)
                yend = np.full_like(ystart, furthest_component.distance.value)

                _add_rays(
                    ax=ax,
                    x=np.stack((xstart, xend), axis=1),
                    y=np.stack((ystart, yend), axis=1),
                    color=source_data.coords["wavelength"].values[inds],
                    cbar=cbar and (i == 0),
                    cmap=cmap,
                    vmin=wmin.value if vmin is None else vmin,
                    vmax=wmax.value if vmax is None else vmax,
                    cax=cax,
                )

            # Plot blocked rays
            inds = rng.choice(
                ids[blocked], size=min(blocked_rays, nblocked), replace=False
            )
            x = np.repeat(
                np.stack(
                    [source_data.coords["birth_time"].values[inds]]
                    + [
                        c.data.coords["toa"]["pulse", i].values[inds]
                        for c in components
                    ],
                    axis=1,
                ),
                repeats,
                axis=1,
            )
            y = np.repeat(
                np.stack(
                    [np.full_like(x[:, 0], source_dist)]
                    + [np.full_like(x[:, 0], c.distance.value) for c in components],
                    axis=1,
                ),
                repeats,
                axis=1,
            )
            for j, c in enumerate(components):
                comp_data = c.data["pulse", i]
                m_others = comp_data.masks["blocked_by_others"].values[inds]
                x[:, 2 * j + 1][m_others] = np.nan
                y[:, 2 * j + 1][m_others] = np.nan
                if "blocked_by_me" in comp_data.masks:
                    m_me = comp_data.masks["blocked_by_me"].values[inds]
                    x[:, 2 * j + 2][m_me] = np.nan
                    y[:, 2 * j + 2][m_me] = np.nan
            _add_rays(ax=ax, x=x, y=y, color="lightgray", zorder=-1)

            # Plot pulse
            time_coord = source_data.coords["birth_time"].values
            tmin = time_coord.min()
            ax.plot([tmin, time_coord.max()], [source_dist] * 2, color="gray", lw=3)
            ax.text(tmin, source_dist, "Pulse", ha="left", va="top", color="gray")
        if furthest_component.toa.data.sum().value > 0:
            toa_max = furthest_component.toa.max().value
        else:
            toa_max = furthest_component.toa.data.coords["toa"].max().value
        dx = 0.05 * toa_max
        # Plot choppers
        for ch in self._choppers.values():
            x0 = ch.open_times.values
            x1 = ch.close_times.values
            x = np.empty(3 * x0.size, dtype=x0.dtype)
            x[0::3] = x0
            x[1::3] = 0.5 * (x0 + x1)
            x[2::3] = x1
            x = np.concatenate(
                ([[0]] if x[0] > 0 else [x[0:1]])
                + [x]
                + ([[toa_max + dx]] if x[-1] < toa_max else [])
            )
            y = np.full_like(x, ch.distance.value)
            y[2::3] = None
            inds = np.argsort(x)
            ax.plot(x[inds], y[inds], color="k")
            ax.text(
                toa_max, ch.distance.value, ch.name, ha="right", va="bottom", color="k"
            )

        # Plot detectors
        for det in self._detectors.values():
            ax.plot([0, toa_max], [det.distance.value] * 2, color="gray", lw=3)
            ax.text(
                0, det.distance.value, det.name, ha="left", va="bottom", color="gray"
            )

        ax.set(xlabel="Time [μs]", ylabel="Distance [m]")
        ax.set_xlim(0 - dx, toa_max + dx)
        if figsize is None:
            inches = fig.get_size_inches()
            fig.set_size_inches((min(inches[0] * self.source.pulses, 12.0), inches[1]))
        fig.tight_layout()
        return Plot(fig=fig, ax=ax)

    def __repr__(self) -> str:
        out = (
            f"Result:\n  Source: {self.source.pulses} pulses, "
            f"{self.source.neutrons} neutrons per pulse.\n  Choppers:\n"
        )
        for name, ch in self._choppers.items():
            out += f"    {name}: {ch._repr_stats()}\n"
        out += "  Detectors:\n"
        for name, det in self._detectors.items():
            out += f"    {name}: {det._repr_stats()}\n"
        return out

    def __str__(self) -> str:
        return self.__repr__()

    def to_nxevent_data(self, key: str | None = None) -> sc.DataArray:
        """
        Convert a detector reading to event data that resembles event data found in a
        NeXus file.

        Parameters
        ----------
        key:
            Name of the detector. If ``None``, all detectors are included.
        """
        start = sc.datetime("2024-01-01T12:00:00.000000")
        period = sc.reciprocal(self.source.frequency)

        keys = list(self._detectors.keys()) if key is None else [key]

        event_data = []
        for name in keys:
            raw_data = self._detectors[name].data.flatten(to="event")
            events = (
                raw_data[~raw_data.masks["blocked_by_others"]]
                .copy()
                .drop_masks("blocked_by_others")
            )
            events.coords["distance"] = sc.broadcast(
                events.coords["distance"], sizes=events.sizes
            ).copy()
            event_data.append(events)

        event_data = sc.concat(event_data, dim=event_data[0].dim)
        dt = period.to(unit=event_data.coords["toa"].unit)
        event_time_zero = (dt * (event_data.coords["toa"] // dt)).to(dtype=int) + start
        event_data.coords["event_time_zero"] = event_time_zero
        event_data.coords["event_time_offset"] = event_data.coords.pop(
            "toa"
        ) % period.to(unit=dt.unit)
        out = (
            event_data.drop_coords(["tof", "speed", "birth_time", "wavelength"])
            .group("distance")
            .rename_dims(distance="detector_number")
        )
        out.coords["Ltotal"] = out.coords.pop("distance")
        return out



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import functools
import warnings
from collections.abc import Callable


class VisibleDeprecationWarning(UserWarning):
    """Visible deprecation warning.

    By default, Python and in particular Jupyter will not show deprecation
    warnings, so this class can be used when a very visible warning is helpful.
    """


VisibleDeprecationWarning.__module__ = 'tof'


def deprecated(message: str = '') -> Callable:
    def decorator(function: Callable) -> Callable:
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f'{function.__name__} is deprecated. {message}',
                VisibleDeprecationWarning,
                stacklevel=2,
            )
            return function(*args, **kwargs)

        return wrapper

    return decorator


__all__ = ['deprecated']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from dataclasses import dataclass, replace
from enum import Enum
from typing import TYPE_CHECKING

import scipp as sc

from .reading import ComponentReading
from .utils import two_pi, var_to_dict

if TYPE_CHECKING:
    try:
        from scippneutron.chopper import DiskChopper
    except ImportError:
        DiskChopper = object


class Direction(Enum):
    CLOCKWISE = 1
    ANTICLOCKWISE = 2


Clockwise = Direction.CLOCKWISE
AntiClockwise = Direction.ANTICLOCKWISE


class Chopper:
    """
    A chopper is a rotating device with cutouts that blocks the beam at certain times.

    Parameters
    ----------
    frequency:
        The frequency of the chopper. Must be positive.
    distance:
        The distance from the source to the chopper.
    name:
        The name of the chopper.
    phase:
        The phase of the chopper. Because the phase offset implemented as a time delay
        on real beamline choppers, it is applied in the opposite direction
        to the chopper rotation direction. For example, if the chopper rotates
        clockwise, a phase of 10 degrees will shift all window angles by 10 degrees
        in the anticlockwise direction, which will result in the windows opening later.
    open:
        The opening angles of the chopper cutouts.
    close:
        The closing angles of the chopper cutouts.
    centers:
        The centers of the chopper cutouts.
    widths:
        The widths of the chopper cutouts.

    Notes
    -----
    Either `open` and `close` or `centers` and `widths` must be provided, but not both.
    """

    def __init__(
        self,
        *,
        frequency: sc.Variable,
        distance: sc.Variable,
        name: str,
        phase: sc.Variable | None = None,
        open: sc.Variable | None = None,
        close: sc.Variable | None = None,
        centers: sc.Variable | None = None,
        widths: sc.Variable | None = None,
        direction: Direction = Clockwise,
    ):
        if frequency <= (0.0 * frequency.unit):
            raise ValueError(f"Chopper frequency must be positive, got {frequency:c}.")
        self.frequency = frequency.to(dtype=float, copy=False)
        if direction not in (Clockwise, AntiClockwise):
            raise ValueError(
                "Chopper direction must be Clockwise or AntiClockwise"
                f", got {direction}."
            )
        if phase is None:
            phase = sc.scalar(0.0, unit='deg')
        self.direction = direction
        # Check that either open/close or centers/widths are provided, but not both
        if tuple(x for x in (open, close, centers, widths) if x is not None) not in (
            (open, close),
            (centers, widths),
        ):
            raise ValueError(
                "Either open/close or centers/widths must be provided, got"
                f" open={open}, close={close}, centers={centers}, widths={widths}."
            )
        if open is None:
            half_width = widths * 0.5
            open = centers - half_width
            close = centers + half_width

        self.open = (open if open.dims else open.flatten(to='cutout')).to(
            dtype=float, copy=False
        )
        self.close = (close if close.dims else close.flatten(to='cutout')).to(
            dtype=float, copy=False
        )
        self.distance = distance.to(dtype=float, copy=False)
        self.phase = phase.to(dtype=float, copy=False)
        self.name = name
        super().__init__()

    @property
    def omega(self) -> sc.Variable:
        """
        The angular velocity of the chopper.
        """
        return two_pi * self.frequency

    def open_close_times(
        self, time_limit: sc.Variable | None = None, unit: str | None = None
    ) -> tuple[sc.Variable, sc.Variable]:
        """
        The times at which the chopper opens and closes.

        Parameters
        ----------
        time_limit:
            Determines how many rotations the chopper needs to perform to reach the time
            limit. If not specified, the chopper will perform a single rotation.
        unit:
            The unit of the returned times. If not specified, the unit of `time_limit`
            is used.
        """
        if time_limit is None:
            time_limit = sc.scalar(0.0, unit='us')
        if unit is None:
            unit = time_limit.unit
        nrot = max(int(sc.ceil((time_limit * self.frequency).to(unit='')).value), 1)
        # We make a unique dim name that is different from self.open.dim and
        # self.close.dim to make use of automatic broadcasting below.
        # We also start at -1 to catch early openings in case the phase or opening
        # angles are large
        phases = sc.arange(
            f"{self.open.dim}-{self.close.dim}", -1, nrot
        ) * two_pi + self.phase.to(unit='rad')

        open_times = self.open.to(unit='rad', copy=False)
        close_times = self.close.to(unit='rad', copy=False)
        # If the chopper is rotating anti-clockwise, we mirror the openings because the
        # first cutout will be the last to open.
        if self.direction == AntiClockwise:
            open_times, close_times = (
                sc.array(
                    dims=close_times.dims,
                    values=(two_pi - close_times).values[::-1],
                    unit=close_times.unit,
                ),
                sc.array(
                    dims=open_times.dims,
                    values=(two_pi - open_times).values[::-1],
                    unit=open_times.unit,
                ),
            )
        # Note that the order is important here: we need (phases + open/close) to get
        # the correct dimension order when we flatten.
        open_times = (phases + open_times).flatten(to=self.open.dim)
        close_times = (phases + close_times).flatten(to=self.close.dim)
        open_times /= self.omega
        close_times /= self.omega
        return (
            open_times.to(unit=unit, copy=False),
            close_times.to(unit=unit, copy=False),
        )

    def __repr__(self) -> str:
        return (
            f"Chopper(name={self.name}, distance={self.distance:c}, "
            f"frequency={self.frequency:c}, phase={self.phase:c}, "
            f"direction={self.direction.name}, cutouts={len(self.open)})"
        )

    def as_dict(self) -> dict:
        """
        Return the chopper as a dictionary.
        """
        return {
            'frequency': self.frequency,
            'open': self.open,
            'close': self.close,
            'distance': self.distance,
            'phase': self.phase,
            'name': self.name,
            'direction': self.direction,
        }

    def as_json(self) -> dict:
        """
        Return the chopper as a JSON-serializable dictionary.
        """
        out = {
            key: var_to_dict(value)
            for key, value in self.as_dict().items()
            if isinstance(value, sc.Variable)
        }
        out.update(
            {
                'type': 'chopper',
                'direction': self.direction.name.lower(),
                'name': self.name,
            }
        )
        return out

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Chopper):
            return NotImplemented
        if self.name != other.name:
            return False
        if self.direction != other.direction:
            return False
        return all(
            sc.identical(getattr(self, field), getattr(other, field))
            for field in ('frequency', 'distance', 'phase', 'open', 'close')
        )

    @classmethod
    def from_diskchopper(
        cls, disk_chopper: DiskChopper, name: str | None = None
    ) -> Chopper:
        """
        Create a Chopper from a scippneutron DiskChopper.

        Example
        -------
        >>> import scipp as sc
        >>> import tof
        >>> from scippneutron.chopper import DiskChopper
        >>> disk_chopper = DiskChopper(
        ...     frequency=sc.scalar(28.0, unit='Hz'),
        ...     beam_position=sc.scalar(30.0, unit='deg'),
        ...     slit_begin=sc.array(
        ...         dims=['slit'], values=[0.0, 90.0, 150.0], unit='deg'
        ...     ),
        ...     slit_end=sc.array(
        ...         dims=['slit'], values=[60.0, 120.0, 210.0], unit='deg'
        ...     ),
        ...     phase=sc.scalar(15.0, unit='deg'),
        ...     axle_position=sc.vector(value=[0.0, 0.0, 13.0], unit='m'),
        ... )
        >>> chopper = tof.Chopper.from_diskchopper(disk_chopper)
        """

        distance = disk_chopper.axle_position.fields.z
        freq = abs(disk_chopper.frequency)
        if name is None:
            name = (
                f"Chopper@{distance.value:.1f}{distance.unit}_"
                f"{int(freq.value)}{freq.unit}"
            )

        return cls(
            frequency=freq,
            direction=AntiClockwise
            if (disk_chopper.frequency.value > 0.0)
            else Clockwise,
            open=disk_chopper.slit_begin - disk_chopper.beam_position,
            close=disk_chopper.slit_end - disk_chopper.beam_position,
            phase=disk_chopper.phase
            if disk_chopper.frequency.value > 0.0
            else -disk_chopper.phase,
            distance=distance,
            name=name,
        )

    @classmethod
    def from_nexus(cls, nexus_chopper, name: str | None = None) -> Chopper:
        """
        Create a Chopper from a NeXus chopper group.

        Example
        -------
        >>> # Assuming a chopper NeXus structure
        >>> import scipp as sc
        >>> import tof
        >>> nexus_chopper = {
        ...     'position': sc.vector([0.0, 0.0, 2.0], unit='m'),
        ...     'rotation_speed': sc.scalar(12.0, unit='Hz'),
        ...     'beam_position': sc.scalar(45.0, unit='deg'),
        ...     'phase': sc.scalar(20.0, unit='deg'),
        ...     'slit_edges': sc.array(
        ...         dims=['slit'],
        ...         values=[0.0, 60.0, 124.0, 126.0],
        ...         unit='deg',
        ...     ),
        ...     'slit_height': sc.scalar(0.4, unit='m'),
        ...     'radius': sc.scalar(0.5, unit='m'),
        ... }
        >>> chopper = tof.Chopper.from_nexus(nexus_chopper)
        """

        distance = nexus_chopper['position'].fields.z
        freq = abs(nexus_chopper['rotation_speed'])
        if name is None:
            name = (
                f"Chopper@{distance.value:.1f}{distance.unit}_"
                f"{int(freq.value)}{freq.unit}"
            )

        return cls(
            frequency=freq,
            direction=AntiClockwise
            if (nexus_chopper['rotation_speed'].value > 0.0)
            else Clockwise,
            open=nexus_chopper['slit_edges'][::2] - nexus_chopper['beam_position'],
            close=nexus_chopper['slit_edges'][1::2] - nexus_chopper['beam_position'],
            phase=nexus_chopper['phase']
            if nexus_chopper['rotation_speed'].value > 0.0
            else -nexus_chopper['phase'],
            distance=distance,
            name=name,
        )

    def to_diskchopper(self) -> DiskChopper:
        """
        Export the chopper as a scippneutron DiskChopper.
        """
        from scippneutron.chopper import DiskChopper

        frequency = (
            self.frequency if self.direction == AntiClockwise else -self.frequency
        )
        phase = self.phase if self.direction == AntiClockwise else -self.phase
        return DiskChopper(
            frequency=frequency,
            beam_position=sc.scalar(0.0, unit='deg'),
            slit_begin=self.open,
            slit_end=self.close,
            phase=phase,
            axle_position=sc.vector(
                value=[0.0, 0.0, self.distance.value], unit=self.distance.unit
            ),
        )


@dataclass(frozen=True)
class ChopperReading(ComponentReading):
    """
    Read-only container for the neutrons that reach the chopper.
    """

    distance: sc.Variable
    name: str
    frequency: sc.Variable
    open: sc.Variable
    close: sc.Variable
    phase: sc.Variable
    open_times: sc.Variable
    close_times: sc.Variable
    data: sc.DataArray

    def _repr_stats(self) -> str:
        return (
            f"visible={int(self.data.sum().value)}, "
            f"blocked={int(self.data.masks['blocked_by_me'].sum().value)}"
        )

    def __repr__(self) -> str:
        return f"""ChopperReading: '{self.name}'
  distance: {self.distance:c}
  frequency: {self.frequency:c}
  phase: {self.phase:c}
  cutouts: {len(self.open)}
  neutrons: {self._repr_stats()}
"""

    def __str__(self) -> str:
        return self.__repr__()

    def __getitem__(self, val: int | slice | tuple[str, int | slice]) -> ChopperReading:
        if isinstance(val, int):
            val = ('pulse', val)
        return replace(self, data=self.data[val])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

from dataclasses import dataclass

import plopp as pp
import scipp as sc

from .utils import Plot


@dataclass(frozen=True)
class ReadingField:
    data: sc.DataArray
    dim: str

    def plot(self, bins: int = 300, **kwargs):
        by_pulse = sc.collapse(self.data, keep="event")
        to_plot = {}
        color = {}
        for key, da in by_pulse.items():
            sel = da[~da.masks["blocked_by_others"]]
            if sel.size == 0:
                continue
            to_plot[key] = sel.hist({self.dim: bins})
            if "blocked_by_me" in self.data.masks:
                name = f"blocked-{key}"
                to_plot[name] = (
                    da[da.masks["blocked_by_me"]]
                    .drop_masks(list(da.masks.keys()))
                    .hist({self.dim: to_plot[key].coords[self.dim]})
                )
                color[name] = "gray"
        if not to_plot:
            raise RuntimeError("Nothing to plot.")
        return pp.plot(to_plot, **{**{"color": color}, **kwargs})

    def min(self):
        da = self.data.copy(deep=False)
        da.data = da.coords[self.dim]
        return da.min().data

    def max(self):
        da = self.data.copy(deep=False)
        da.data = da.coords[self.dim]
        return da.max().data

    def __repr__(self) -> str:
        da = self.data.copy(deep=False)
        da.data = da.coords[self.dim]
        return (
            f"{self.dim}: min={da.min().data:c}, max={da.max().data:c}, "
            f"events={int(self.data.sum().value)}"
        )

    def __str__(self) -> str:
        return self.__repr__()

    def __getitem__(self, val: int | slice | tuple[str, int | slice]) -> ReadingField:
        if isinstance(val, int):
            val = ('pulse', val)
        return self.__class__(data=self.data[val], dim=self.dim)


def _make_reading_field(da: sc.DataArray, dim: str) -> ReadingField:
    return ReadingField(
        data=sc.DataArray(
            data=da.data,
            coords={dim: da.coords[dim]},
            masks=da.masks,
        ),
        dim=dim,
    )


class ComponentReading:
    """
    Data reading for a component placed in the beam path. The reading will have a
    record of the arrival times and wavelengths of the neutrons that passed through it.
    """

    @property
    def toa(self) -> ReadingField:
        """
        Time of arrival of the neutrons at the component.
        """
        return _make_reading_field(self.data, dim="toa")

    @property
    def eto(self) -> ReadingField:
        """
        Event time offset of the neutrons at the component (= toa modulo pulse period).
        """
        return _make_reading_field(self.data, dim="eto")

    @property
    def wavelength(self) -> ReadingField:
        """
        Wavelength of the neutrons at the component.
        """
        return _make_reading_field(self.data, dim="wavelength")

    @property
    def birth_time(self) -> ReadingField:
        """
        Birth time of the neutrons at the source.
        """
        return _make_reading_field(self.data, dim="birth_time")

    @property
    def speed(self) -> ReadingField:
        """
        Speed of the neutrons at the component.
        """
        return _make_reading_field(self.data, dim="speed")

    def plot(self, bins: int = 300) -> Plot:
        """
        Plot both the toa and wavelength data side by side.

        Parameters
        ----------
        bins:
            Number of bins to use for histogramming the neutrons.
        """
        return self.toa.plot(bins=bins) + self.wavelength.plot(bins=bins)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""
Pre-configured chopper and detector parameters for Dream.
"""

from ...chopper import Chopper
from ...detector import Detector
from ...model import make_beamline

dream_high_flux = {
    "PSC1": {
        "type": "chopper",
        "frequency": {"value": 14.0, "unit": "Hz"},
        "phase": {"value": 286.0 - 180.0, "unit": "deg"},
        "distance": {"value": 6.145, "unit": "m"},
        "centers": {
            "value": [0, 72, 86.4, 115.2, 172.8, 273.6, 288.0, 302.4],
            "unit": "deg",
        },
        "widths": {
            "value": [2.46, 3.02, 3.27, 3.27, 5.02, 3.93, 3.93, 2.46],
            "unit": "deg",
        },
        "direction": "anti-clockwise",
    },
    "PSC2": {
        "type": "chopper",
        "frequency": {"value": 14.0, "unit": "Hz"},
        "phase": {"value": 236.0, "unit": "deg"},
        "distance": {"value": 6.155, "unit": "m"},
        "centers": {
            "value": [0, 28.8, 57.6, 144.0, 158.4, 216.0, 259.2, 316.8],
            "unit": "deg",
        },
        "widths": {
            "value": [2.46, 3.60, 3.60, 3.23, 3.27, 3.77, 3.94, 2.62],
            "unit": "deg",
        },
        "direction": "clockwise",
    },
    "OC": {
        "type": "chopper",
        "frequency": {"value": 14.0, "unit": "Hz"},
        "phase": {"value": 297.0 - 180.0 - 90.0, "unit": "deg"},
        "distance": {"value": 6.174, "unit": "m"},
        "centers": {"value": [0.0], "unit": "deg"},
        "widths": {"value": [27.6], "unit": "deg"},
        "direction": "anti-clockwise",
    },
    "BC": {
        "type": "chopper",
        "frequency": {"value": 112.0, "unit": "Hz"},
        "phase": {"value": 240.0 - 180.0, "unit": "deg"},
        "distance": {"value": 9.78, "unit": "m"},
        "centers": {"value": [0.0, 180.0], "unit": "deg"},
        "widths": {"value": [73.75, 73.75], "unit": "deg"},
        "direction": "anti-clockwise",
    },
    "T0": {
        "type": "chopper",
        "frequency": {"value": 28.0, "unit": "Hz"},
        "phase": {"value": 280.0 - 180.0, "unit": "deg"},
        "distance": {"value": 13.05, "unit": "m"},
        "centers": {"value": [0.0], "unit": "deg"},
        "widths": {"value": [314.9], "unit": "deg"},
        "direction": "anti-clockwise",
    },
    "mantle": {"type": "detector", "distance": {"value": 77.675, "unit": "m"}},
    "end-cap": {"type": "detector", "distance": {"value": 77.675, "unit": "m"}},
    "high-resolution": {"type": "detector", "distance": {"value": 79.05, "unit": "m"}},
    "sans": {"type": "detector", "distance": {"value": 79.05, "unit": "m"}},
}


def dream(
    high_flux=False, high_resolution=False
) -> dict[str, list[Chopper] | list[Detector]]:
    if high_flux and high_resolution:
        raise ValueError("Select either high flux or high resolution, not both.")
    if high_flux:
        return make_beamline(dream_high_flux)
    if high_resolution:
        # TODO: Add high-resolution configuration
        raise NotImplementedError("High-resolution configuration not yet implemented.")
    raise ValueError("Either high_flux or high_resolution must be True.")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

import tof


@pytest.fixture
def make_chopper():
    def _make_chopper(topen, tclose, f, phase, distance, name):
        two_pi = 2.0 * np.pi * sc.units.rad
        aopen = two_pi * sc.concat(topen, dim='cutout').to(unit='s') * f
        aclose = two_pi * sc.concat(tclose, dim='cutout').to(unit='s') * f
        return tof.Chopper(
            frequency=f,
            open=aopen,
            close=aclose,
            phase=phase,
            distance=distance,
            name=name,
        )

    return _make_chopper


@pytest.fixture
def make_source():
    def _make_source(arrival_times, distance, pulses=1, frequency=None):
        # Arrival times are distance * m_over_h * wavelength
        return tof.Source.from_neutrons(
            birth_times=sc.array(
                dims=['event'], values=[0.0] * len(arrival_times), unit='s'
            ),
            wavelengths=arrival_times.to(unit='s') / (distance * tof.utils.m_over_h),
            pulses=pulses,
            frequency=frequency,
        )

    return _make_source


@pytest.fixture
def dummy_chopper():
    return tof.Chopper(
        frequency=sc.scalar(1.0, unit="Hz"),
        open=sc.array(dims=['cutout'], values=[0.0], unit='deg'),
        close=sc.array(dims=['cutout'], values=[1.0], unit='deg'),
        phase=sc.scalar(0.0, unit="deg"),
        distance=sc.scalar(1.0, unit="m"),
        name='dummy_chopper',
    )


@pytest.fixture
def dummy_detector():
    return tof.Detector(
        distance=sc.scalar(1.0, unit="m"),
        name='dummy_detector',
    )


@pytest.fixture
def dummy_source():
    return tof.Source.from_neutrons(
        birth_times=sc.array(dims=['event'], values=[0.0], unit='s'),
        wavelengths=sc.array(dims=['event'], values=[1.0], unit='angstrom'),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from dataclasses import dataclass, replace

import scipp as sc

from .reading import ComponentReading
from .utils import var_to_dict


class Detector:
    """
    A detector component does not block any neutrons, it sees all neutrons passing
    through it.

    Parameters
    ----------
    distance:
        The distance from the source to the detector.
    name:
        The name of the detector.
    """

    def __init__(self, distance: sc.Variable, name: str):
        self.distance = distance.to(dtype=float, copy=False)
        self.name = name

    def __repr__(self) -> str:
        return f"Detector(name={self.name}, distance={self.distance:c})"

    def as_dict(self) -> dict:
        """
        Return the detector as a dictionary.
        """
        return {'distance': self.distance, 'name': self.name}

    def as_json(self) -> dict:
        """
        Return the detector as a JSON-serializable dictionary.
        """
        return {
            'type': 'detector',
            'distance': var_to_dict(self.distance),
            'name': self.name,
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Detector):
            return NotImplemented
        return self.name == other.name and sc.identical(self.distance, other.distance)


@dataclass(frozen=True)
class DetectorReading(ComponentReading):
    """
    Read-only container for the neutrons that reach the detector.
    """

    distance: sc.Variable
    name: str
    data: sc.DataArray

    def _repr_stats(self) -> str:
        return f"visible={int(self.data.sum().value)}"

    def __repr__(self) -> str:
        return f"""DetectorReading: '{self.name}'
  distance: {self.distance:c}
  neutrons: {self._repr_stats()}
"""

    def __str__(self) -> str:
        return self.__repr__()

    def __getitem__(
        self, val: int | slice | tuple[str, int | slice]
    ) -> DetectorReading:
        if isinstance(val, int):
            val = ('pulse', val)
        return replace(self, data=self.data[val])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


import numpy as np
import pytest
import scipp as sc

import tof


def test_ess_pulse():
    source = tof.Source(facility='ess', neutrons=100_000)
    # Check that the time distribution is low on edges and high in the middle
    times = source.data['pulse', 0].hist(birth_time=300)
    mean = times.mean()
    assert (times[0] < 0.5 * mean).value
    assert (times[-1] < 0.5 * mean).value
    assert (times[150] > 1.5 * mean).value
    # Check that there are more neutrons at low wavelengths
    wavs = source.data['pulse', 0].hist(wavelength=300)
    assert (wavs[:150].sum() > 1.5 * wavs[150:].sum()).value


@pytest.mark.parametrize('facility', ['ESS', 'Ess'])
def test_ess_pulse_uppercase(facility):
    source = tof.Source(facility=facility, neutrons=100_000)
    assert source.neutrons == 100_000
    assert sc.identical(source.frequency, sc.scalar(14.0, unit="Hz"))
    assert source.pulses == 1
    assert source.facility == 'ess'


def test_creation_from_neutrons():
    birth_times = sc.array(dims=['event'], values=[1000.0, 1500.0, 2000.0], unit='us')
    wavelengths = sc.array(dims=['event'], values=[1.0, 5.0, 10.0], unit='angstrom')
    source = tof.Source.from_neutrons(
        birth_times=birth_times,
        wavelengths=wavelengths,
    )
    assert source.neutrons == 3
    assert sc.identical(source.data['pulse', 0].coords['birth_time'], birth_times)
    assert sc.identical(source.data['pulse', 0].coords['wavelength'], wavelengths)


def test_creation_from_distribution_flat():
    birth_time = sc.linspace('birth_time', 1.0, 3.0, 100, unit='ms')
    p_time = sc.DataArray(
        data=sc.ones(sizes=birth_time.sizes),
        coords={'birth_time': birth_time},
    )
    wavelength = sc.linspace('wavelength', 1.0, 10.0, 100, unit='angstrom')
    p_wav = sc.DataArray(
        data=sc.ones(sizes=wavelength.sizes),
        coords={'wavelength': wavelength},
    )

    N = 123456
    source = tof.Source.from_distribution(
        neutrons=N,
        p_time=p_time,
        p_wav=p_wav,
    )

    assert source.neutrons == N
    assert source.data.sizes['event'] == N
    h = source.data['pulse', 0].hist(birth_time=10)
    assert sc.allclose(
        h.data,
        sc.full(value=N / 10.0, sizes={'birth_time': 10}, unit='counts'),
        rtol=sc.scalar(0.05),
    )


@pytest.mark.parametrize('distribution_2d', [False, True])
def test_creation_from_distribution(distribution_2d):
    v = np.ones(90) * 0.1
    v[30:60] = 1.0

    p_time = sc.DataArray(
        data=sc.array(dims=['birth_time'], values=v),
        coords={
            'birth_time': sc.linspace('birth_time', 0.0, 8000.0, len(v), unit='us')
        },
    )
    p_wav = sc.DataArray(
        data=sc.linspace('wavelength', 1.0, 4.0, 100, unit='angstrom'),
        coords={
            'wavelength': sc.linspace('wavelength', 1.0, 4.0, 100, unit='angstrom')
        },
    )
    if distribution_2d:
        source = tof.Source.from_distribution(neutrons=100_000, p=p_wav * p_time)
    else:
        source = tof.Source.from_distribution(
            neutrons=100_000, p_time=p_time, p_wav=p_wav
        )

    assert source.neutrons == 100_000
    da = source.data['pulse', 0]
    assert da.hist(
        birth_time=sc.array(dims=['birth_time'], values=[2100.0, 2900.0], unit='us')
    ).data.sum() > sc.scalar(0.0, unit='counts')
    assert da.hist(
        birth_time=sc.array(dims=['birth_time'], values=[5100.0, 5900.0], unit='us')
    ).data.sum() > sc.scalar(0.0, unit='counts')

    left = da.hist(
        birth_time=sc.array(dims=['birth_time'], values=[0.0, 2000.0], unit='us')
    ).data.sum()
    mid = da.hist(
        birth_time=sc.array(dims=['birth_time'], values=[3000.0, 5000.0], unit='us')
    ).data.sum()
    right = da.hist(
        birth_time=sc.array(dims=['birth_time'], values=[6000.0, 8000.0], unit='us')
    ).data.sum()
    rtol = sc.scalar(0.05)
    assert sc.isclose(mid / left, sc.scalar(10.0), rtol=rtol)
    assert sc.isclose(mid / right, sc.scalar(10.0), rtol=rtol)

    # Make sure distribution is monotonically increasing
    h = da.hist(wavelength=10)
    diff = h.data[1:] - h.data[:-1]
    assert sc.all(diff > sc.scalar(0.0, unit='counts'))


def test_non_integer_neutrons():
    source = tof.Source(facility='ess', neutrons=1e5)
    assert source.neutrons == 100_000


def test_multiple_pulses_ess():
    source = tof.Source(facility='ess', neutrons=100_000, pulses=3)
    assert source.data.sizes['pulse'] == 3
    assert source.data.sizes['event'] == 100_000
    assert sc.identical(source.frequency, sc.scalar(14.0, unit='Hz'))
    bins = (sc.arange('birth_time', 4) / source.frequency).to(unit='us')
    h = source.data.flatten(to='event').hist(birth_time=bins)
    assert sc.allclose(
        h.data, sc.full(value=1e5, sizes={'birth_time': 3}, unit='counts')
    )


def test_multiple_pulses_from_distribution():
    v = np.ones(9) * 0.1
    v[3:6] = 1.0

    p_time = sc.DataArray(
        data=sc.array(dims=['birth_time'], values=v),
        coords={
            'birth_time': sc.linspace('birth_time', 0.0, 8000.0, len(v), unit='us')
        },
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0], unit='angstrom'
            )
        },
    )

    source = tof.Source.from_distribution(
        neutrons=123_987,
        p_time=p_time,
        p_wav=p_wav,
        pulses=2,
        frequency=sc.scalar(100.0, unit='Hz'),
    )
    assert source.data.sizes['pulse'] == 2
    assert source.data.sizes['event'] == 123_987
    bins = (sc.arange('birth_time', 3) / source.frequency).to(unit='us')
    h = source.data.flatten(to='event').hist(birth_time=bins)
    assert sc.allclose(
        h.data, sc.full(value=123987.0, sizes={'birth_time': 2}, unit='counts')
    )


def test_multiple_pulses_from_neutrons():
    birth_times = sc.array(
        dims=['event'], values=[1111.0, 1567.0, 856.0, 2735.0], unit='us'
    )
    wavelengths = sc.array(
        dims=['event'], values=[1.0, 5.0, 8.0, 10.0], unit='angstrom'
    )
    source = tof.Source.from_neutrons(
        birth_times=birth_times,
        wavelengths=wavelengths,
        pulses=3,
        frequency=sc.scalar(50.0, unit='Hz'),
    )
    assert source.data.sizes['pulse'] == 3
    assert source.data.sizes['event'] == 4
    offsets = (sc.arange('pulse', 3) / source.frequency).to(unit='us')
    assert sc.allclose(source.data.coords['birth_time'], birth_times + offsets)
    assert sc.allclose(
        source.data.coords['wavelength'],
        sc.broadcast(wavelengths, sizes={'pulse': 3, 'event': 4}),
    )


def test_source_length():
    N = 17
    source = tof.Source(facility='ess', neutrons=3124, pulses=N)
    assert len(source) == N


def test_multiple_pulses_from_distribution_no_frequency_raises():
    v = np.ones(9) * 0.1
    v[3:6] = 1.0
    p_time = sc.DataArray(
        data=sc.array(dims=['birth_time'], values=v),
        coords={'birth_time': sc.linspace('birth_time', 0.0, 8.0, len(v), unit='ms')},
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0], unit='angstrom'
            )
        },
    )
    with pytest.raises(
        ValueError, match='If pulses is greater than one, a frequency must be supplied.'
    ):
        tof.Source.from_distribution(
            neutrons=22696,
            p_time=p_time,
            p_wav=p_wav,
            pulses=3,
        )


def test_multiple_pulses_from_neutrons_no_frequency_raises():
    birth_times = sc.array(
        dims=['event'], values=[1111.0, 1567.0, 856.0, 2735.0], unit='us'
    )
    wavelengths = sc.array(
        dims=['event'], values=[1.0, 5.0, 8.0, 10.0], unit='angstrom'
    )
    with pytest.raises(
        ValueError, match='If pulses is greater than one, a frequency must be supplied.'
    ):
        tof.Source.from_neutrons(
            birth_times=birth_times,
            wavelengths=wavelengths,
            pulses=3,
        )


def test_source_repr_does_not_raise():
    assert repr(tof.Source(facility='ess', neutrons=100_000)) is not None
    assert repr(tof.Source(facility='ess', neutrons=100_000, pulses=3)) is not None


def test_seed_ess_pulse():
    a = tof.Source(facility='ess', neutrons=100_000, seed=1234)
    b = tof.Source(facility='ess', neutrons=100_000, seed=1234)
    assert sc.identical(a.data, b.data)
    c = tof.Source(facility='ess', neutrons=100_000, seed=0)
    assert not sc.identical(a.data, c.data)


def test_seed_from_distribution():
    v = np.ones(9) * 0.1
    v[3:6] = 1.0

    p_time = sc.DataArray(
        data=sc.array(dims=['birth_time'], values=v),
        coords={
            'birth_time': sc.linspace('birth_time', 0.0, 8000.0, len(v), unit='us')
        },
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[1.0, 2.0, 3.0, 4.0], unit='angstrom'
            )
        },
    )

    a = tof.Source.from_distribution(
        neutrons=100_000, p_time=p_time, p_wav=p_wav, seed=12
    )
    b = tof.Source.from_distribution(
        neutrons=100_000, p_time=p_time, p_wav=p_wav, seed=12
    )
    assert sc.identical(a.data, b.data)
    c = tof.Source.from_distribution(
        neutrons=100_000, p_time=p_time, p_wav=p_wav, seed=1
    )
    assert not sc.identical(a.data, c.data)


def test_source_from_distribution_with_one_element_raises():
    p_time = sc.DataArray(
        data=sc.array(dims=['birth_time'], values=[1.0]),
        coords={'birth_time': sc.array(dims=['birth_time'], values=[1.0], unit='ms')},
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[2.0]),
        coords={
            'wavelength': sc.array(dims=['wavelength'], values=[2.0], unit='angstrom')
        },
    )

    with pytest.raises(
        ValueError,
        match='Distribution must have at least 2 points in each dimension',
    ):
        tof.Source.from_distribution(neutrons=10, p_time=p_time, p_wav=p_wav)


def test_source_from_distrbution_all_zero_probability_raises():
    # All zero p_time
    p_time = sc.DataArray(
        data=sc.array(dims=['birth_time'], values=[0.0, 0.0, 0.0]),
        coords={'birth_time': sc.linspace('birth_time', 0.0, 2.0, 3, unit='ms')},
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[0.0, 1.0, 0.0]),
        coords={'wavelength': sc.linspace('wavelength', 1.0, 3.0, 3, unit='angstrom')},
    )

    with pytest.raises(
        ValueError,
        match='Time distribution must have at least one positive probability value.',
    ):
        tof.Source.from_distribution(neutrons=10, p_time=p_time, p_wav=p_wav)

    # All zero p_wav
    p_time = sc.DataArray(
        data=sc.array(dims=['birth_time'], values=[0.0, 1.0, 0.0]),
        coords={'birth_time': sc.linspace('birth_time', 0.0, 2.0, 3, unit='ms')},
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[0.0, 0.0, 0.0]),
        coords={'wavelength': sc.linspace('wavelength', 1.0, 3.0, 3, unit='angstrom')},
    )

    with pytest.raises(
        ValueError,
        match=(
            'Wavelength distribution must have at least one positive probability value'
        ),
    ):
        tof.Source.from_distribution(neutrons=10, p_time=p_time, p_wav=p_wav)

    # All zero p_time and p_wav
    p_time = sc.DataArray(
        data=sc.array(dims=['birth_time'], values=[0.0, 0.0, 0.0]),
        coords={'birth_time': sc.linspace('birth_time', 0.0, 2.0, 3, unit='ms')},
    )
    p_wav = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[0.0, 0.0, 0.0]),
        coords={'wavelength': sc.linspace('wavelength', 1.0, 3.0, 3, unit='angstrom')},
    )

    with pytest.raises(
        ValueError,
        match='Distribution must have at least one positive probability value.',
    ):
        tof.Source.from_distribution(neutrons=10, p=p_wav * p_time)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import json
import os
import tempfile

import numpy as np
import pytest
import scipp as sc

import tof

Hz = sc.Unit('Hz')
deg = sc.Unit('deg')
meter = sc.Unit('m')
ms = sc.Unit('ms')


def test_one_chopper_one_opening(make_chopper, make_source):
    # Make a chopper open from 10-20 ms. Assume zero phase.
    topen = 10.0 * ms
    tclose = 20.0 * ms
    chopper = make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper',
    )
    detector = tof.Detector(distance=20 * meter, name='detector')

    # Make a pulse with 3 neutrons with one neutron going through the chopper opening
    # and the other two neutrons on either side of the opening.
    source = make_source(
        arrival_times=sc.concat(
            [0.9 * topen, 0.5 * (topen + tclose), 1.1 * tclose], dim='event'
        ),
        distance=chopper.distance,
    )

    model = tof.Model(source=source, choppers=[chopper], detectors=[detector])
    res = model.run()

    toa = res.choppers['chopper'].toa.data
    assert toa.sum().value == 1
    assert toa.masks['blocked_by_me'].sum().value == 2
    assert np.array_equal(
        toa.masks['blocked_by_me'].squeeze().values, [True, False, True]
    )
    assert sc.allclose(
        toa.coords['toa']['pulse', 0],
        (
            source.data.coords['wavelength']['pulse', 0]
            * chopper.distance
            * tof.utils.m_over_h
        ).to(unit='us'),
    )

    toa = res.detectors['detector'].toa.data
    assert toa.sum().value == 1
    assert toa.masks['blocked_by_others'].sum().value == 2
    assert sc.allclose(
        toa.coords['toa']['pulse', 0],
        (
            source.data.coords['wavelength']['pulse', 0]
            * detector.distance
            * tof.utils.m_over_h
        ).to(unit='us'),
    )


def test_two_choppers_one_opening(make_chopper, make_source):
    # Make a first chopper open from 5-16 ms. Assume zero phase.
    topen = 5.0 * ms
    tclose = 16.0 * ms
    chopper1 = make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )

    # Make a second chopper open from 15-20 ms.
    chopper2 = make_chopper(
        topen=[15.0 * ms],
        tclose=[20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )

    detector = tof.Detector(distance=20 * meter, name='detector')

    # Make a pulse with 3 neutrons with 2 neutrons going through the first chopper
    # opening and only one making it through the second chopper.
    source = make_source(
        arrival_times=sc.concat(
            [1.5 * topen, 0.5 * (topen + tclose), 1.1 * tclose], dim='event'
        ),
        distance=chopper1.distance,
    )

    model = tof.Model(
        source=source, choppers=[chopper1, chopper2], detectors=[detector]
    )
    res = model.run()

    ch1_toas = res.choppers['chopper1'].toa.data
    ch2_toas = res.choppers['chopper2'].toa.data
    det = res.detectors['detector'].toa.data

    # Blocks the third neutron
    assert ch1_toas.sum().value == 2
    assert ch1_toas.masks['blocked_by_me'].sum().value == 1
    assert ch1_toas.masks['blocked_by_others'].sum().value == 0
    assert np.array_equal(
        ch1_toas.masks['blocked_by_me'].squeeze().values, [False, False, True]
    )
    assert sc.allclose(
        ch1_toas.coords['toa']['pulse', 0],
        (
            source.data.coords['wavelength']['pulse', 0]
            * chopper1.distance
            * tof.utils.m_over_h
        ).to(unit='us'),
    )

    # Blocks the first neutron
    assert ch2_toas.sum().value == 1
    assert ch2_toas.masks['blocked_by_me'].sum().value == 1
    assert ch2_toas.masks['blocked_by_others'].sum().value == 1
    assert np.array_equal(
        ch2_toas.masks['blocked_by_me'].squeeze().values, [True, False, False]
    )
    assert sc.allclose(
        ch2_toas.coords['toa']['pulse', 0],
        (
            source.data.coords['wavelength']['pulse', 0]
            * chopper2.distance
            * tof.utils.m_over_h
        ).to(unit='us'),
    )

    assert det.sum().value == 1
    assert det.masks['blocked_by_others'].sum().value == 2
    assert sc.allclose(
        det.coords['toa']['pulse', 0],
        (
            source.data.coords['wavelength']['pulse', 0]
            * detector.distance
            * tof.utils.m_over_h
        ).to(unit='us'),
    )


def test_two_choppers_one_and_two_openings(make_chopper, make_source):
    topen = 5.0 * ms
    tclose = 16.0 * ms
    chopper1 = make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )

    chopper2 = make_chopper(
        topen=[9.0 * ms, 15.0 * ms],
        tclose=[12.0 * ms, 20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )

    detector = tof.Detector(distance=20 * meter, name='detector')

    # Make a pulse with 7 neutrons:
    # - 2 neutrons blocked by chopper1
    # - 3 neutrons blocked by chopper2
    # - 2 neutrons detected (one through each of chopper2's openings)
    source = make_source(
        arrival_times=sc.concat(
            [
                0.9 * topen,
                1.01 * topen,
                1.3 * topen,
                1.8 * topen,
                0.5 * (topen + tclose),
                0.85 * tclose,
                1.1 * tclose,
            ],
            dim='event',
        ),
        distance=chopper1.distance,
    )

    model = tof.Model(
        source=source, choppers=[chopper1, chopper2], detectors=[detector]
    )
    res = model.run()

    assert res.choppers['chopper1'].toa.data.sum().value == 5
    assert res.choppers['chopper1'].toa.data.masks['blocked_by_me'].sum().value == 2
    assert res.choppers['chopper2'].toa.data.sum().value == 2
    assert res.choppers['chopper2'].toa.data.masks['blocked_by_me'].sum().value == 3
    assert res.detectors['detector'].toa.data.sum().value == 2
    assert (
        res.detectors['detector'].toa.data.masks['blocked_by_others'].sum().value == 5
    )


def test_neutron_conservation(make_chopper):
    N = 100_000
    source = tof.Source(facility='ess', neutrons=N)

    chopper1 = make_chopper(
        topen=[5.0 * ms],
        tclose=[16.0 * ms],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )
    chopper2 = make_chopper(
        topen=[9.0 * ms, 15.0 * ms],
        tclose=[15.0 * ms, 20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )

    detector = tof.Detector(distance=20 * meter, name='detector')
    model = tof.Model(
        source=source, choppers=[chopper1, chopper2], detectors=[detector]
    )
    res = model.run()

    ch1 = res.choppers['chopper1'].toa.data
    ch2 = res.choppers['chopper2'].toa.data
    det = res.detectors['detector'].toa.data

    assert ch1.sizes['event'] == N
    assert ch2.sizes['event'] == N

    assert sc.identical(ch1.masks['blocked_by_me'], ch2.masks['blocked_by_others'])
    assert sc.identical(
        det.masks['blocked_by_others'],
        ch2.masks['blocked_by_me'] | ch1.masks['blocked_by_me'],
    )
    assert det.sum().value + det.masks['blocked_by_others'].sum().value == N


def test_source_not_at_origin(make_chopper):
    N = 100_000
    source1 = tof.Source(facility='ess', neutrons=N, seed=123)

    chopper1 = make_chopper(
        topen=[5.0 * ms],
        tclose=[16.0 * ms],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )
    chopper2 = make_chopper(
        topen=[9.0 * ms, 15.0 * ms],
        tclose=[15.0 * ms, 20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )

    detector = tof.Detector(distance=20 * meter, name='detector')
    model1 = tof.Model(
        source=source1, choppers=[chopper1, chopper2], detectors=[detector]
    )
    res1 = model1.run()

    source2 = tof.Source(facility='ess', neutrons=N, seed=123)
    source2._distance = 7.0 * meter
    model2 = tof.Model(
        source=source2, choppers=[chopper1, chopper2], detectors=[detector]
    )
    res2 = model2.run()

    # Less neutrons make it through when the source is closer to the choppers
    assert res1['detector'].data.sum().value > res2['detector'].data.sum().value

    # Less fast neutrons make it through when the source is closer
    assert (
        res2['detector'].wavelength.min().value
        > res1['detector'].wavelength.min().value
    )
    # More slow neutrons make it through when the source is closer
    assert (
        res2['detector'].wavelength.max().value
        > res1['detector'].wavelength.max().value
    )


def test_add_chopper_and_detector(dummy_chopper, dummy_detector, dummy_source):
    chopper = dummy_chopper
    detector = dummy_detector
    model = tof.Model(source=dummy_source)
    model.add(chopper)
    assert 'dummy_chopper' in model.choppers
    model.add(detector)
    assert 'dummy_detector' in model.detectors


def test_add_components_with_same_name_raises(
    dummy_chopper, dummy_detector, dummy_source
):
    chopper = dummy_chopper
    detector = dummy_detector
    model = tof.Model(source=dummy_source)
    model.add(chopper)
    with pytest.raises(
        KeyError, match='Component with name dummy_chopper already exists'
    ):
        model.add(chopper)
    model.add(detector)
    with pytest.raises(
        KeyError, match='Component with name dummy_detector already exists'
    ):
        model.add(detector)
    detector2 = tof.Detector(distance=22 * meter, name='dummy_chopper')
    with pytest.raises(
        KeyError, match='Component with name dummy_chopper already exists'
    ):
        model.add(detector2)


def test_create_model_with_duplicate_component_names_raises(
    dummy_chopper, dummy_detector, dummy_source
):
    chopper = dummy_chopper
    detector = dummy_detector
    with pytest.raises(
        ValueError, match="More than one component named 'dummy_chopper' found"
    ):
        tof.Model(source=dummy_source, choppers=[chopper, chopper])
    with pytest.raises(
        ValueError, match="More than one component named 'dummy_detector' found"
    ):
        tof.Model(source=dummy_source, detectors=[detector, detector])


def test_iter(dummy_chopper, dummy_detector, dummy_source):
    chopper = dummy_chopper
    detector = dummy_detector
    model = tof.Model(source=dummy_source)
    model.add(chopper)
    assert 'dummy_chopper' in model.choppers
    model.add(detector)
    assert 'dummy_detector' in model.detectors


def test_remove(dummy_chopper, dummy_detector, dummy_source):
    chopper = dummy_chopper
    detector = dummy_detector
    model = tof.Model(source=dummy_source, choppers=[chopper], detectors=[detector])
    del model['dummy_chopper']
    assert 'dummy_chopper' not in model
    assert 'dummy_detector' in model
    del model['dummy_detector']
    assert 'dummy_detector' not in model


def test_getitem(dummy_chopper, dummy_detector, dummy_source):
    chopper = dummy_chopper
    detector = dummy_detector
    model = tof.Model(source=dummy_source, choppers=[chopper], detectors=[detector])
    assert model['dummy_chopper'] is chopper
    assert model['dummy_detector'] is detector
    with pytest.raises(KeyError, match='No component with name foo'):
        model['foo']


def test_input_can_be_single_component(dummy_chopper, dummy_detector, dummy_source):
    chopper = dummy_chopper
    detector = dummy_detector
    model = tof.Model(source=dummy_source, choppers=chopper, detectors=detector)
    assert 'dummy_chopper' in model.choppers
    assert 'dummy_detector' in model.detectors


def test_bad_input_type_raises(dummy_chopper, dummy_detector, dummy_source):
    chopper = dummy_chopper
    detector = dummy_detector
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source, choppers='bad chopper')
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source, choppers=[chopper], detectors='abc')
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source, choppers=[chopper, 'bad chopper'])
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source, detectors=(1234, detector))
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source, choppers=[detector])
    with pytest.raises(TypeError, match='Invalid input type'):
        _ = tof.Model(source=dummy_source, detectors=[chopper])


def test_model_repr_does_not_raise(make_chopper):
    N = 10_000
    source = tof.Source(facility='ess', neutrons=N)
    chopper1 = make_chopper(
        topen=[5.0 * ms],
        tclose=[16.0 * ms],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper1',
    )
    chopper2 = make_chopper(
        topen=[9.0 * ms, 15.0 * ms],
        tclose=[15.0 * ms, 20.0 * ms],
        f=15.0 * Hz,
        phase=0.0 * deg,
        distance=15 * meter,
        name='chopper2',
    )
    detector = tof.Detector(distance=20 * meter, name='detector')
    model = tof.Model(
        source=source, choppers=[chopper1, chopper2], detectors=[detector]
    )
    assert repr(model) is not None


def test_component_distance(make_chopper, make_source):
    # Make a chopper open from 10-20 ms. Assume zero phase.
    topen = 10.0 * ms
    tclose = 20.0 * ms
    chopper = make_chopper(
        topen=[topen],
        tclose=[tclose],
        f=10.0 * Hz,
        phase=0.0 * deg,
        distance=10 * meter,
        name='chopper',
    )
    monitor = tof.Detector(distance=17 * meter, name='monitor')
    detector = tof.Detector(distance=20 * meter, name='detector')

    # Make a pulse with 3 neutrons with one neutron going through the chopper opening
    # and the other two neutrons on either side of the opening.
    source = make_source(
        arrival_times=sc.concat(
            [0.9 * topen, 0.5 * (topen + tclose), 1.1 * tclose], dim='event'
        ),
        distance=chopper.distance,
    )

    model = tof.Model(source=source, choppers=[chopper], detectors=[monitor, detector])
    res = model.run()

    assert sc.identical(res['monitor'].data.coords['distance'], monitor.distance)
    assert sc.identical(res['detector'].data.coords['distance'], detector.distance)
    assert sc.identical(res['chopper'].data.coords['distance'], chopper.distance)


def test_to_nxevent_data():
    source = tof.Source(facility='ess', neutrons=100_000)
    choppers = [
        tof.Chopper(
            frequency=14.0 * Hz,
            open=sc.array(dims=['cutout'], values=[0.0], unit='deg'),
            close=sc.array(dims=['cutout'], values=[10.0], unit='deg'),
            phase=90.0 * deg,
            distance=8.0 * meter,
            name="chopper",
        )
    ]
    detectors = [
        tof.Detector(distance=26.0 * meter, name='monitor'),
        tof.Detector(distance=32.0 * meter, name='detector'),
    ]
    model = tof.Model(source=source, choppers=choppers, detectors=detectors)
    res = model.run()

    # There should be 1 pulse for monitor data, and 2 pulses for detector data as it
    # wraps around the pulse period.
    for key, npulses in zip(('monitor', 'detector'), (1, 2), strict=True):
        nxevent_data = res.to_nxevent_data(key)
        assert sc.identical(res['monitor'].data.sum().data, nxevent_data.sum().data)
        grouped = nxevent_data.group('event_time_zero')
        assert grouped.sizes['event_time_zero'] == npulses
        assert nxevent_data.bins.concat().value.coords[
            'event_time_offset'
        ].min() >= sc.scalar(0.0, unit='us')
        assert nxevent_data.bins.concat().value.coords[
            'event_time_offset'
        ].max() <= sc.reciprocal(source.frequency).to(unit='us')

    # Test when we include all detectors at once
    nxevent_data = res.to_nxevent_data()
    assert nxevent_data.sizes == {'detector_number': 2}


def test_model_from_json():
    params = {
        "source": {"type": "source", "facility": "ess", "neutrons": 1e4, "pulses": 1},
        "chopper1": {
            "type": "chopper",
            "frequency": {"value": 56.0, "unit": "Hz"},
            "phase": {"value": 93.244, "unit": "deg"},
            "distance": {"value": 6.85, "unit": "m"},
            "open": {
                "value": [-1.9419, 49.5756, 98.9315, 146.2165, 191.5176, 234.9179],
                "unit": "deg",
            },
            "close": {
                "value": [1.9419, 55.7157, 107.2332, 156.5891, 203.8741, 249.1752],
                "unit": "deg",
            },
            "direction": "clockwise",
        },
        "chopper2": {
            "type": "chopper",
            "frequency": {"value": 14.0, "unit": "Hz"},
            "phase": {"value": 31.080 + 14.0, "unit": "deg"},
            "distance": {"value": 8.45, "unit": "m"},
            "open": {"value": [-23.6029], "unit": "deg"},
            "close": {"value": [23.6029], "unit": "deg"},
            "direction": "clockwise",
        },
        "detector": {"type": "detector", "distance": {"value": 60.5, "unit": "m"}},
    }

    with tempfile.TemporaryDirectory() as path:
        fname = os.path.join(path, "my_instrument.json")
        with open(fname, "w") as f:
            json.dump(params, f)
        model = tof.Model.from_json(fname)

    assert 'chopper1' in model.choppers
    assert 'chopper2' in model.choppers
    assert 'detector' in model.detectors
    assert model.source.facility == 'ess'
    assert model.source.neutrons == 10_000
    assert model.source.pulses == 1
    results = model.run()
    assert 'chopper1' in results.choppers
    assert 'chopper2' in results.choppers
    assert 'detector' in results.detectors


def test_model_from_json_no_source():
    params = {
        "chopper1": {
            "type": "chopper",
            "frequency": {"value": 56.0, "unit": "Hz"},
            "phase": {"value": 93.244, "unit": "deg"},
            "distance": {"value": 6.85, "unit": "m"},
            "open": {
                "value": [-1.9419, 49.5756, 98.9315, 146.2165, 191.5176, 234.9179],
                "unit": "deg",
            },
            "close": {
                "value": [1.9419, 55.7157, 107.2332, 156.5891, 203.8741, 249.1752],
                "unit": "deg",
            },
            "direction": "clockwise",
        },
        "chopper2": {
            "type": "chopper",
            "frequency": {"value": 14.0, "unit": "Hz"},
            "phase": {"value": 45.080, "unit": "deg"},
            "distance": {"value": 8.45, "unit": "m"},
            "open": {"value": [-23.6029], "unit": "deg"},
            "close": {"value": [23.6029], "unit": "deg"},
            "direction": "clockwise",
        },
        "detector": {"type": "detector", "distance": {"value": 60.5, "unit": "m"}},
    }

    with tempfile.TemporaryDirectory() as path:
        fname = os.path.join(path, "my_instrument.json")
        with open(fname, "w") as f:
            json.dump(params, f)
        model = tof.Model.from_json(fname)

    assert 'chopper1' in model.choppers
    assert 'chopper2' in model.choppers
    assert 'detector' in model.detectors
    assert model.source is None
    with pytest.raises(ValueError, match="No source has been defined for this model"):
        _ = model.run()

    model.source = tof.Source(facility='ess', neutrons=10_000, pulses=1)
    results = model.run()
    assert 'chopper1' in results.choppers
    assert 'chopper2' in results.choppers
    assert 'detector' in results.detectors


def test_as_json():
    source = tof.Source(facility='ess', neutrons=100_000, pulses=1, seed=42)
    choppers = [
        tof.Chopper(
            frequency=14.0 * Hz,
            open=sc.array(dims=['cutout'], values=[0.1], unit='deg'),
            close=sc.array(dims=['cutout'], values=[12.0], unit='deg'),
            phase=88.0 * deg,
            distance=9.5 * meter,
            name="chopperZ",
        )
    ]
    detectors = [
        tof.Detector(distance=26.0 * meter, name='monitor'),
        tof.Detector(distance=32.0 * meter, name='detector'),
    ]
    model = tof.Model(source=source, choppers=choppers, detectors=detectors)
    json_dict = model.as_json()

    json_source = json_dict['source']
    assert json_source['facility'] == 'ess'
    assert json_source['neutrons'] == 100_000
    assert json_source['pulses'] == 1
    assert json_source['seed'] == 42
    assert json_source['type'] == 'source'

    json_chop = json_dict['chopperZ']
    assert json_chop['distance']['value'] == 9.5
    assert json_chop['distance']['unit'] == 'm'
    assert json_chop['frequency']['value'] == 14.0
    assert json_chop['frequency']['unit'] == 'Hz'
    assert json_chop['phase']['value'] == 88.0
    assert json_chop['phase']['unit'] == 'deg'
    assert json_chop['open']['value'] == [0.1]
    assert json_chop['open']['unit'] == 'deg'
    assert json_chop['close']['value'] == [12.0]
    assert json_chop['close']['unit'] == 'deg'
    assert json_chop['direction'] == 'clockwise'
    assert json_chop['type'] == 'chopper'

    json_det1 = json_dict['monitor']
    assert json_det1['distance']['value'] == 26.0
    assert json_det1['distance']['unit'] == 'm'
    assert json_det1['type'] == 'detector'

    json_det2 = json_dict['detector']
    assert json_det2['distance']['value'] == 32.0
    assert json_det2['distance']['unit'] == 'm'
    assert json_det2['type'] == 'detector'


def test_model_to_json_roundtrip():
    source = tof.Source(facility='ess', neutrons=100_000, pulses=1, seed=42)
    choppers = [
        tof.Chopper(
            frequency=14.0 * Hz,
            open=sc.array(dims=['cutout'], values=[0.1], unit='deg'),
            close=sc.array(dims=['cutout'], values=[12.0], unit='deg'),
            phase=88.0 * deg,
            distance=9.5 * meter,
            name="chopperZ",
        )
    ]
    detectors = [
        tof.Detector(distance=26.0 * meter, name='monitor'),
        tof.Detector(distance=32.0 * meter, name='detector'),
    ]
    original = tof.Model(source=source, choppers=choppers, detectors=detectors)

    with tempfile.TemporaryDirectory() as path:
        fname = os.path.join(path, "my_instrument.json")
        original.to_json(fname)
        loaded = tof.Model.from_json(fname)

    assert source.facility == loaded.source.facility
    assert source.neutrons == loaded.source.neutrons
    assert source.pulses == loaded.source.pulses
    assert source.seed == loaded.source.seed

    for ch in choppers:
        assert ch == loaded.choppers[ch.name]
    for det in detectors:
        assert det == loaded.detectors[det.name]


def test_model_run_without_source_raises(dummy_chopper, dummy_detector):
    chopper = dummy_chopper
    detector = dummy_detector
    model = tof.Model(choppers=[chopper], detectors=[detector])
    with pytest.raises(ValueError, match="No source has been defined for this model"):
        model.run()


def test_model_run_without_components_raises(dummy_source):
    source = dummy_source
    model = tof.Model(source=source)
    with pytest.raises(
        ValueError, match="Cannot run model: no components have been defined."
    ):
        model.run()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

import scipp as sc

import tof


def test_as_json():
    detector = tof.Detector(distance=sc.scalar(54.0, unit='m'), name='DetectorX')

    json_str = detector.as_json()

    assert json_str['type'] == 'detector'
    assert json_str['distance']['value'] == detector.distance.value
    assert json_str['distance']['unit'] == str(detector.distance.unit)
    assert json_str['name'] == detector.name


def test_equal():
    detector1 = tof.Detector(distance=sc.scalar(54.0, unit='m'), name='DetectorX')
    detector2 = tof.Detector(distance=sc.scalar(54.0, unit='m'), name='DetectorX')
    detector3 = tof.Detector(distance=sc.scalar(60.0, unit='m'), name='DetectorY')
    detector4 = tof.Detector(distance=sc.scalar(54.0, unit='m'), name='DetectorY')
    detector5 = tof.Detector(
        distance=sc.scalar(54.0, unit='m').to(unit='cm'), name='DetectorX'
    )

    assert detector1 == detector2
    assert detector1 != detector3
    assert detector4 != detector3
    assert detector1 != detector4
    assert detector1 != detector5



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from __future__ import annotations

import warnings
from itertools import chain

import scipp as sc

from .chopper import AntiClockwise, Chopper, Clockwise
from .detector import Detector
from .result import Result
from .source import Source

ComponentType = Chopper | Detector


def _input_to_dict(
    obj: None | list[ComponentType] | tuple[ComponentType, ...] | ComponentType,
    kind: type,
):
    if isinstance(obj, list | tuple):
        out = {}
        for item in obj:
            new = _input_to_dict(item, kind=kind)
            for key in new.keys():
                if key in out:
                    raise ValueError(f"More than one component named '{key}' found.")
            out.update(new)
        return out
    elif isinstance(obj, kind):
        return {obj.name: obj}
    elif obj is None:
        return {}
    else:
        raise TypeError(
            "Invalid input type. Must be a Chopper or a Detector, "
            "or a list/tuple of Choppers or Detectors."
        )


def _array_or_none(container: dict, key: str) -> sc.Variable | None:
    return (
        sc.array(
            dims=["cutout"], values=container[key]["value"], unit=container[key]["unit"]
        )
        if key in container
        else None
    )


def make_beamline(instrument: dict) -> dict[str, list[Chopper] | list[Detector]]:
    """
    Create choppers and detectors from a dictionary.
    The dictionary is typically loaded from a pre-configured instrument library, or
    from a JSON file.

    Parameters
    ----------
    instrument:
        A dictionary defining the instrument components.
        Each key is the name of a component, and the value is a dictionary with the
        component parameters. Each component dictionary must have a "type" key, which
        must be either "chopper" or "detector". Other keys depend on the component
        type, see the documentation of the :class:`Chopper` and :class:`Detector`
        classes for details.
    """
    choppers = []
    detectors = []
    for name, comp in instrument.items():
        if comp["type"] == "chopper":
            direction = comp["direction"].lower()
            if direction == "clockwise":
                _dir = Clockwise
            elif any(x in direction for x in ("anti", "counter")):
                _dir = AntiClockwise
            else:
                raise ValueError(
                    f"Chopper direction must be 'clockwise' or 'anti-clockwise', got "
                    f"'{comp['direction']}' for component {name}."
                )
            choppers.append(
                Chopper(
                    frequency=comp["frequency"]["value"]
                    * sc.Unit(comp["frequency"]["unit"]),
                    direction=_dir,
                    open=_array_or_none(comp, "open"),
                    close=_array_or_none(comp, "close"),
                    centers=_array_or_none(comp, "centers"),
                    widths=_array_or_none(comp, "widths"),
                    phase=comp["phase"]["value"] * sc.Unit(comp["phase"]["unit"]),
                    distance=comp["distance"]["value"]
                    * sc.Unit(comp["distance"]["unit"]),
                    name=name,
                )
            )
        elif comp["type"] == "detector":
            detectors.append(
                Detector(
                    distance=comp["distance"]["value"]
                    * sc.Unit(comp["distance"]["unit"]),
                    name=name,
                )
            )
        elif comp["type"] == "source":
            continue
        else:
            raise ValueError(
                f"Unknown component type: {comp['type']} for component {name}. "
                "Supported types are 'chopper', 'detector', and 'source'."
            )
    return {"choppers": choppers, "detectors": detectors}


class Model:
    """
    A class that represents a neutron instrument.
    It is defined by a list of choppers, a list of detectors, and a source.

    Parameters
    ----------
    choppers:
        A list of choppers.
    detectors:
        A list of detectors.
    source:
        A source of neutrons.
    """

    def __init__(
        self,
        source: Source | None = None,
        choppers: Chopper | list[Chopper] | tuple[Chopper, ...] | None = None,
        detectors: Detector | list[Detector] | tuple[Detector, ...] | None = None,
    ):
        self.choppers = _input_to_dict(choppers, kind=Chopper)
        self.detectors = _input_to_dict(detectors, kind=Detector)
        self.source = source

    @classmethod
    def from_json(cls, filename: str) -> Model:
        """
        Create a model from a JSON file.

        Currently, only sources from facilities are supported when loading from JSON.

        Parameters
        ----------
        filename:
            The path to the JSON file.
        """
        import json

        with open(filename) as f:
            instrument = json.load(f)
        beamline = make_beamline(instrument)
        source = None
        for item in instrument.values():
            if item.get("type") == "source":
                if "facility" not in item:
                    raise ValueError(
                        "Currently, only sources from facilities are supported when "
                        "loading from JSON."
                    )
                source_args = item.copy()
                del source_args["type"]
                source = Source(**source_args)
                break
        return cls(source=source, **beamline)

    def as_json(self) -> dict:
        """
        Return the model as a JSON-serializable dictionary.
        If the source is not from a facility, it is not included in the output.
        """
        instrument_dict = {}
        if self.source is not None:
            if self.source.facility is None:
                warnings.warn(
                    "The source is not from a facility, so it will not be included in "
                    "the JSON output.",
                    UserWarning,
                    stacklevel=2,
                )
            else:
                instrument_dict['source'] = self.source.as_json()
        for ch in self.choppers.values():
            instrument_dict[ch.name] = ch.as_json()
        for det in self.detectors.values():
            instrument_dict[det.name] = det.as_json()
        return instrument_dict

    def to_json(self, filename: str):
        """
        Save the model to a JSON file.
        If the source is not from a facility, it is not included in the output.

        Parameters
        ----------
        filename:
            The path to the JSON file.
        """
        import json

        with open(filename, 'w') as f:
            json.dump(self.as_json(), f, indent=2)

    def add(self, component):
        """
        Add a component to the instrument.
        Component names must be unique across choppers and detectors.

        Parameters
        ----------
        component:
            A chopper or detector.
        """
        if component.name in chain(self.choppers, self.detectors):
            raise KeyError(
                f"Component with name {component.name} already exists. "
                "If you wish to replace/update an existing component, use "
                "``model.choppers['name'] = new_chopper`` or "
                "``model.detectors['name'] = new_detector``."
            )
        if isinstance(component, Chopper):
            self.choppers[component.name] = component
        elif isinstance(component, Detector):
            self.detectors[component.name] = component
        else:
            raise TypeError(
                f"Cannot add component of type {type(component)} to the model."
            )

    def remove(self, name: str):
        """
        Remove a component.

        Parameters
        ----------
        name:
            The name of the component to remove.
        """
        if name in self.choppers:
            del self.choppers[name]
        elif name in self.detectors:
            del self.detectors[name]
        else:
            raise KeyError(f"No component with name {name} was found.")

    def __iter__(self):
        return chain(self.choppers, self.detectors)

    def __getitem__(self, name) -> Chopper | Detector:
        if name not in self:
            raise KeyError(f"No component with name {name} was found.")
        return self.choppers[name] if name in self.choppers else self.detectors[name]

    def __delitem__(self, name):
        self.remove(name)

    def run(self):
        """
        Run the simulation.
        """
        if self.source is None:
            raise ValueError(
                "No source has been defined for this model. Please add a source using "
                "`model.source = Source(...)` before running the simulation."
            )
        components = sorted(
            chain(self.choppers.values(), self.detectors.values()),
            key=lambda c: c.distance.value,
        )

        if len(components) == 0:
            raise ValueError("Cannot run model: no components have been defined.")

        if components[0].distance < self.source.distance:
            raise ValueError(
                "At least one component is located before the source "
                "itself. Please check the distances of the components."
            )

        birth_time = self.source.data.coords['birth_time']
        speed = self.source.data.coords['speed']
        initial_mask = sc.ones(sizes=birth_time.sizes, unit=None, dtype=bool)

        result_choppers = {}
        result_detectors = {}
        time_limit = (
            birth_time
            + ((components[-1].distance - self.source.distance) / speed).to(
                unit=birth_time.unit
            )
        ).max()
        for c in components:
            container = result_detectors if isinstance(c, Detector) else result_choppers
            container[c.name] = c.as_dict()
            container[c.name]['data'] = self.source.data.copy(deep=False)
            t = birth_time + ((c.distance - self.source.distance) / speed).to(
                unit=birth_time.unit, copy=False
            )
            container[c.name]['data'].coords['toa'] = t
            container[c.name]['data'].coords['eto'] = t % (
                1 / self.source.frequency
            ).to(unit=t.unit, copy=False)
            container[c.name]['data'].coords['distance'] = c.distance
            # TODO: remove 'tof' coordinate once deprecation period is over
            container[c.name]['data'].coords['tof'] = t
            if isinstance(c, Detector):
                container[c.name]['data'].masks['blocked_by_others'] = ~initial_mask
                continue
            m = sc.zeros(sizes=t.sizes, unit=None, dtype=bool)
            to, tc = c.open_close_times(time_limit=time_limit)
            container[c.name].update({'open_times': to, 'close_times': tc})
            for i in range(len(to)):
                m |= (t > to[i]) & (t < tc[i])
            combined = initial_mask & m
            container[c.name]['data'].masks['blocked_by_others'] = ~initial_mask
            container[c.name]['data'].masks['blocked_by_me'] = ~m & initial_mask
            initial_mask = combined

        return Result(
            source=self.source, choppers=result_choppers, detectors=result_detectors
        )

    def __repr__(self) -> str:
        out = f"Model:\n  Source: {self.source}\n  Choppers:\n"
        for name, ch in self.choppers.items():
            out += f"    {name}: {ch}\n"
        out += "  Detectors:\n"
        for name, det in self.detectors.items():
            out += f"    {name}: {det}\n"
        return out

    def __str__(self) -> str:
        return self.__repr__()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from .dream import dream
from .odin import odin

sources = {
    "ess": {"path": "ess/ess.h5", "hash": "md5:f382e4a5c2171b4ef2285fb6625544e4"},
    "ess-odin": {
        "path": "ess/ess-odin.h5",
        "hash": "md5:4d64102d2a743bd7672967ad0cb94872",
    },
}


__all__ = ["dream", "odin", "sources"]



# SPDX-License-Identifier: BSD-3-Clause

import uuid
from collections.abc import Callable
from functools import partial, reduce
from itertools import chain
from typing import Any

import ipywidgets as ipw
import matplotlib.pyplot as plt
import scipp as sc
from matplotlib.backend_bases import PickEvent
from mpl_toolkits.axes_grid1 import make_axes_locatable

from . import facilities
from .chopper import AntiClockwise, Chopper, Clockwise
from .detector import Detector
from .model import Model
from .source import Source


class ChopperWidget(ipw.VBox):
    def __init__(self):
        self._uid = uuid.uuid4()
        self.frequency_widget = ipw.FloatText(
            description="Frequency", continuous_update=True
        )
        self.open_widget = ipw.Text(description="Open")
        self.close_widget = ipw.Text(description="Close")
        self.phase_widget = ipw.FloatText(description="Phase", continuous_update=True)
        self.distance_widget = ipw.FloatText(
            description="Distance", continuous_update=True
        )
        self.name_widget = ipw.Text(description="Name")
        self.direction_widget = ipw.Dropdown(
            options=["clockwise", "anti-clockwise"], description="Direction"
        )
        self.enabled_widget = ipw.Checkbox(
            value=True, indent=False, layout={"width": "100px"}
        )
        self.delete_widget = ipw.Button(
            icon="trash-alt", tooltip="Delete chopper", layout={"width": "40px"}
        )
        super().__init__(
            [
                self.name_widget,
                self.frequency_widget,
                self.open_widget,
                self.close_widget,
                self.phase_widget,
                self.distance_widget,
                self.direction_widget,
                ipw.HBox(
                    [
                        ipw.Label(value="Enabled"),
                        self.enabled_widget,
                        self.delete_widget,
                    ]
                ),
            ],
            layout={
                "border": "1px solid lightgray",
                "display": "flex",
                "flex_flow": "column",
                "align_items": "flex-end",
            },
        )

    def continuous_update(self, callback: Callable):
        self.frequency_widget.observe(callback, names="value")
        self.phase_widget.observe(callback, names="value")
        self.distance_widget.observe(callback, names="value")
        self.open_widget.observe(callback, names="value")
        self.close_widget.observe(callback, names="value")
        self.direction_widget.observe(callback, names="value")
        self.enabled_widget.observe(callback, names="value")


class DetectorWidget(ipw.VBox):
    def __init__(self):
        self._uid = uuid.uuid4()
        self.distance_widget = ipw.FloatText(
            description="Distance", continuous_update=True
        )
        self.name_widget = ipw.Text(description="Name")
        self.enabled_widget = ipw.Checkbox(
            value=True, indent=False, layout={"width": "100px"}
        )
        self.delete_widget = ipw.Button(
            icon="trash-alt", tooltip="Delete detector", layout={"width": "40px"}
        )
        super().__init__(
            [
                self.name_widget,
                self.distance_widget,
                ipw.HBox(
                    [
                        ipw.Label(value="Enabled"),
                        self.enabled_widget,
                        self.delete_widget,
                    ]
                ),
            ],
            layout={
                "border": "1px solid lightgray",
                "display": "flex",
                "flex_flow": "column",
                "align_items": "flex-end",
            },
        )

    def continuous_update(self, callback: Callable):
        self.distance_widget.observe(callback, names="value")
        self.enabled_widget.observe(callback, names="value")


class SetupWidget(ipw.VBox):
    def __init__(self):
        self.facility_widget = ipw.Dropdown(
            options=["ess", "ess-odin"], description="Facility"
        )
        self.neutrons_widget = ipw.IntText(value=100_000, description="Neutrons")
        self.pulses_widget = ipw.IntText(value=1, description="Pulses")
        self.distance_widget = ipw.FloatText(
            description="Distance", continuous_update=True
        )
        self.instrument_widget = ipw.Dropdown(
            options=[
                "ESS: Odin",
                "ESS: Odin (pulse-skipping)",
                "ESS: Dream (high-flux)",
            ],
            description="Instrument",
            value=None,
        )
        self.facility_widget.observe(self._update_source_distance, names="value")
        super().__init__(
            [
                ipw.Label(value="Source:"),
                self.facility_widget,
                self.neutrons_widget,
                self.pulses_widget,
                self.distance_widget,
                ipw.HTML("<hr>"),
                self.instrument_widget,
            ]
        )

    def _update_source_distance(self, _: dict):
        source = Source(facility=self.facility_widget.value, neutrons=1, pulses=1)
        self.distance_widget.value = source.distance.to(unit='m').value

    def continuous_update(self, callback: Callable):
        self.facility_widget.observe(callback, names="value")
        self.neutrons_widget.observe(callback, names="value")
        self.pulses_widget.observe(callback, names="value")
        self.distance_widget.observe(callback, names="value")


INSTRUMENT_LIBRARY = {
    "ESS: Odin": facilities.ess.odin(pulse_skipping=False),
    "ESS: Odin (pulse-skipping)": facilities.ess.odin(pulse_skipping=True),
    "ESS: Dream (high-flux)": facilities.ess.dream(high_flux=True),
}


class TofWidget:
    def __init__(self):
        self.top_bar = ipw.HBox()

        self.run_button = ipw.Button(description="Run")
        self.continuous_update = ipw.Checkbox(
            description="Continuous update",
            value=False,
            indent=False,
            layout={"width": "180px"},
        )
        self.visible_rays = ipw.IntText(
            description="Visible rays",
            value=1000,
            continuous_update=True,
            layout={"width": "170px"},
        )
        self.blocked_rays = ipw.IntText(
            description="Blocked rays",
            value=0,
            continuous_update=True,
            layout={"width": "170px"},
        )
        with plt.ioff():
            self.time_distance_fig, self.time_distance_ax = plt.subplots(figsize=(8, 6))
            divider = make_axes_locatable(self.time_distance_ax)
            self.time_distance_cax = divider.append_axes("right", "4%", pad="5%")
            self.toa_wav_fig, self.toa_wav_ax = plt.subplots(1, 2, figsize=(11.5, 3.75))
            self.time_distance_fig.canvas.header_visible = False
            self.toa_wav_fig.canvas.header_visible = False
            self._legend_pick_connection = None
            self.toa_wav_fig.canvas.toolbar.toggle_yscale = self.toggle_yscale
            self.toa_wav_fig.canvas.toolbar.toolitems = [
                *self.toa_wav_fig.canvas.toolbar.toolitems,
                ("Logy", "Toggle log scale", "arrows-alt-v", "toggle_yscale"),
            ]

        self.setup_widget = SetupWidget()
        self.setup_widget.continuous_update(self.maybe_update)
        self.setup_widget.instrument_widget.observe(
            self.populate_from_instrument, names="value"
        )

        self.choppers_container = ipw.Accordion()
        self.add_chopper_button = ipw.Button(description="Add chopper")
        self.add_chopper_button.on_click(self.add_chopper)
        self.choppers_widget = ipw.VBox(
            [self.add_chopper_button, self.choppers_container]
        )

        self.detectors_container = ipw.Accordion()
        self.add_detector_button = ipw.Button(description="Add detector")
        self.add_detector_button.on_click(self.add_detector)
        self.detectors_widget = ipw.VBox(
            [self.add_detector_button, self.detectors_container]
        )

        tab_contents = ["Setup", "Choppers", "Detectors"]
        children = [self.setup_widget, self.choppers_widget, self.detectors_widget]
        self.tab = ipw.Tab(layout={"height": "650px", "width": "374px"})
        self.tab.children = children
        self.tab.titles = tab_contents
        self.top_bar.children = [
            ipw.VBox(
                [
                    ipw.HBox(
                        [
                            self.run_button,
                            self.continuous_update,
                            self.visible_rays,
                            self.blocked_rays,
                        ]
                    ),
                    self.time_distance_fig.canvas,
                ]
            ),
            self.tab,
        ]

        self.choppers_widget.children[0].on_click(self.add_chopper)
        self.detectors_widget.children[0].on_click(self.add_detector)
        self.run_button.on_click(self.run)
        self.visible_rays.observe(self.plot_time_distance, names="value")
        self.blocked_rays.observe(self.plot_time_distance, names="value")

        self.main_widget = ipw.VBox([self.top_bar, self.toa_wav_fig.canvas])

    def populate_from_instrument(self, change):
        cont_update_value = self.continuous_update.value
        self.continuous_update.value = False
        for ch in self.choppers_container.children:
            self.remove_chopper(None, uid=ch._uid)
        for det in self.detectors_container.children:
            self.remove_detector(None, uid=det._uid)
        params = INSTRUMENT_LIBRARY[change["new"]]
        for ch in params["choppers"]:
            self.add_chopper(None)
            chop = self.choppers_container.children[-1]
            chop.frequency_widget.value = ch.frequency.to(unit='Hz').value
            chop.open_widget.value = ", ".join(
                str(x) for x in ch.open.to(unit='deg').values
            )
            chop.close_widget.value = ", ".join(
                str(x) for x in ch.close.to(unit='deg').values
            )
            chop.phase_widget.value = ch.phase.to(unit='deg').value
            chop.distance_widget.value = ch.distance.to(unit='m').value
            chop.name_widget.value = ch.name
        for d in params["detectors"]:
            self.add_detector(None)
            det = self.detectors_container.children[-1]
            det.distance_widget.value = d.distance.to(unit='m').value
            det.name_widget.value = d.name
        self.run(None)
        self.continuous_update.value = cont_update_value

    def toggle_yscale(self):
        scale = "log" if self.toa_wav_ax[0].get_yscale() == "linear" else "linear"
        self.toa_wav_ax[0].set_yscale(scale)
        self.toa_wav_ax[1].set_yscale(scale)

    def maybe_update(self, _: Any):
        if self.continuous_update.value:
            self.run(None)
        else:
            self.run_button.style = {"button_color": "lightgreen"}

    def sync_chopper_titles(self, _: Any):
        self.choppers_container.titles = tuple(
            f"{c.name_widget.value} ({int(c.frequency_widget.value)}Hz - "
            f"{c.distance_widget.value:.1f}m)"
            for c in self.choppers_container.children
        )

    def update_tab_titles(self):
        self.tab.titles = [
            self.tab.titles[0],
            f"Choppers ({len(self.choppers_container.children)})",
            f"Detectors ({len(self.detectors_container.children)})",
        ]

    def add_chopper(self, _: Any):
        new_chopper = ChopperWidget()
        new_chopper.name_widget.observe(self.sync_chopper_titles)
        new_chopper.frequency_widget.observe(self.sync_chopper_titles, names="value")
        new_chopper.distance_widget.observe(self.sync_chopper_titles, names="value")
        new_chopper.delete_widget.on_click(
            partial(self.remove_chopper, uid=new_chopper._uid)
        )
        new_chopper.continuous_update(self.maybe_update)
        children = (*self.choppers_container.children, new_chopper)
        self.choppers_container.children = children
        self.choppers_container.selected_index = len(children) - 1
        self.update_tab_titles()

    def remove_chopper(self, _: Any, uid: str):
        self.choppers_container.children = tuple(
            c for c in self.choppers_container.children if c._uid != uid
        )
        self.update_tab_titles()
        self.maybe_update(None)

    def sync_detector_titles(self, _: Any):
        self.detectors_container.titles = tuple(
            f"{d.name_widget.value} ({d.distance_widget.value:.1f}m)"
            for d in self.detectors_container.children
        )

    def add_detector(self, _: Any):
        new_detector = DetectorWidget()
        new_detector.name_widget.observe(self.sync_detector_titles, names="value")
        new_detector.distance_widget.observe(self.sync_detector_titles, names="value")
        new_detector.delete_widget.on_click(
            partial(self.remove_detector, uid=new_detector._uid)
        )
        new_detector.continuous_update(self.maybe_update)
        children = (*self.detectors_container.children, new_detector)
        self.detectors_container.children = children
        self.detectors_container.selected_index = len(children) - 1
        self.update_tab_titles()

    def remove_detector(self, _: Any, uid: str):
        self.detectors_container.children = tuple(
            d for d in self.detectors_container.children if d._uid != uid
        )
        self.update_tab_titles()
        self.maybe_update(None)

    def plot_time_distance(self, _: Any | None = None):
        self.time_distance_ax.clear()
        self.time_distance_cax.clear()
        self.results.plot(
            ax=self.time_distance_ax,
            cax=self.time_distance_cax,
            visible_rays=self.visible_rays.value,
            blocked_rays=self.blocked_rays.value,
        )
        self.time_distance_fig.tight_layout()

    def run(self, _: Any):
        source = Source(
            facility=self.setup_widget.facility_widget.value,
            neutrons=int(self.setup_widget.neutrons_widget.value),
            pulses=int(self.setup_widget.pulses_widget.value),
        )
        source._distance = sc.scalar(self.setup_widget.distance_widget.value, unit="m")

        choppers = [
            Chopper(
                frequency=sc.scalar(ch.frequency_widget.value, unit="Hz"),
                open=sc.array(
                    dims=['cutout'],
                    values=[
                        float(v) for v in ch.open_widget.value.replace(",", " ").split()
                    ],
                    unit="deg",
                ),
                close=sc.array(
                    dims=['cutout'],
                    values=[
                        float(v)
                        for v in ch.close_widget.value.replace(",", " ").split()
                    ],
                    unit="deg",
                ),
                phase=sc.scalar(ch.phase_widget.value, unit="deg"),
                distance=sc.scalar(ch.distance_widget.value, unit="m"),
                name=ch.name_widget.value,
                direction={"clockwise": Clockwise, "anti-clockwise": AntiClockwise}[
                    ch.direction_widget.value
                ],
            )
            for ch in self.choppers_container.children
            if ch.enabled_widget.value
        ]

        detectors = [
            Detector(
                distance=sc.scalar(det.distance_widget.value, unit="m"),
                name=det.name_widget.value,
            )
            for det in self.detectors_container.children
            if det.enabled_widget.value
        ]

        model = Model(source=source, choppers=choppers, detectors=detectors)
        self.results = model.run()
        self.plot_time_distance()

        components = sorted(
            chain(self.results.choppers.values(), self.results.detectors.values()),
            key=lambda c: c.distance,
        )

        self.toa_wav_ax[0].clear()
        self.toa_wav_ax[1].clear()
        if self._legend_pick_connection is not None:
            self.toa_wav_fig.canvas.mpl_disconnect(self._legend_pick_connection)
        for p in range(source.pulses):
            for i, c in enumerate(components):
                label = c.name if p == 0 else None
                da = c.data["pulse", p]
                sel = ~reduce(lambda x, y: x | y, (da.masks[m] for m in da.masks))
                subset = da[sel]
                _, _, toa_patches = self.toa_wav_ax[0].hist(
                    subset.coords["toa"].values,
                    bins=300,
                    histtype="step",
                    lw=1.5,
                    color=f"C{i}",
                    label=label,
                )
                _, _, wav_patches = self.toa_wav_ax[1].hist(
                    subset.coords["wavelength"].values,
                    bins=300,
                    histtype="step",
                    lw=1.5,
                    color=f"C{i}",
                    label=label,
                )
                for patch in chain(toa_patches, wav_patches):
                    patch._component_id = i

        self.toa_wav_ax[0].set(xlabel="Time-of-arrival [μs]", ylabel="Counts")
        self.toa_wav_ax[1].set(xlabel="Wavelength [Å]", ylabel="Counts")
        self.toa_legend = self.toa_wav_ax[0].legend()
        self.wav_legend = self.toa_wav_ax[1].legend()
        self.toa_wav_fig.tight_layout()

        # Clickable legend
        self.map_legend_to_ax = {}
        for i, (toa_patch, wav_patch) in enumerate(
            zip(
                self.toa_legend.get_patches(),
                self.wav_legend.get_patches(),
                strict=True,
            )
        ):
            toa_patch.set_picker(5)
            wav_patch.set_picker(5)
            self.map_legend_to_ax[toa_patch] = i
            self.map_legend_to_ax[wav_patch] = i
        self._legend_pick_connection = self.toa_wav_fig.canvas.mpl_connect(
            "pick_event", self.on_legend_pick
        )

        self.run_button.style = {"button_color": "#eeeeee"}

    def on_legend_pick(self, event: PickEvent):
        legend_patch = event.artist
        # Do nothing if the source of the event is not a legend line.
        if legend_patch not in self.map_legend_to_ax:
            return
        ind = self.map_legend_to_ax[legend_patch]
        for ax in self.toa_wav_ax:
            for patch in ax.patches:
                if patch._component_id == ind:
                    visible = not patch.get_visible()
                    patch.set_visible(visible)
        alpha = 1.0 if visible else 0.2
        self.toa_legend.get_patches()[ind].set_alpha(alpha)
        self.wav_legend.get_patches()[ind].set_alpha(alpha)
        self.toa_wav_fig.canvas.draw_idle()


def Dashboard():
    """
    An interactive app that allows to add choppers and detectors in a widget-powered
    dashboard.
    """
    w = TofWidget()
    w.setup_widget.instrument_widget.value = "ESS: Odin (pulse-skipping)"
    return w.main_widget



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401, I

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"


import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=['facilities'],
    submod_attrs={
        'chopper': ['AntiClockwise', 'Chopper', 'ChopperReading', 'Clockwise'],
        'dashboard': ['Dashboard'],
        'detector': ['Detector', 'DetectorReading'],
        'model': ['Model'],
        'reading': ['ComponentReading', 'ReadingField'],
        'result': ['Result'],
        'source': ['Source', 'SourceParameters'],
    },
)

del importlib
del lazy
